const Chunk = @import("./script/Chunk.zig");

const Table = @import("./script/Table.zig");

const debug = @import("./debug.zig");

const file = @import("./file.zig");

const hashes = @import("./hashes.zig");

const io = @import("./io.zig");

const map = @import("./map.zig");

const scalars = @import("./scalars.zig");

const stack = @import("./stack.zig");

const std = @import("std");

const tokens = @import("./script/tokens.zig");

const tree = @import("./script/tree.zig");

const utf8 = @import("./utf8.zig");

///
/// Archetypes of errors that may occur in the virtual machine during run-time.
///
pub const Error = std.mem.Allocator.Error || io.Error || error {
	IllegalState,
	TypeMismatch,
	BadOperation,
	BadSyntax,
};

///
/// Fixed-length integer number type.
///
pub const Fixed = i32;

///
/// Floating point real number type.
///
pub const Float = f64;

///
/// Supertype for all numeric types.
///
pub const Numeric = union (enum) {
	fixed: Fixed,
	float: Float,
	vector2: Vector2,
	vector3: Vector3,
};

///
/// Runtime-polymorphic data type for representing all data types supported by the virtual machine.
///
pub const Object = opaque {
	const Internal = struct {
		ref_count: u16,

		payload: union (enum) {
			false,
			true,
			float: Float,
			fixed: Fixed,
			symbol: [*:0]const u8,
			vector2: Vector2,
			vector3: Vector3,
			syscall: *const Syscall,
			boxed: ?*Object,

			string: struct {
				ptr: [*]u8,
				len: Fixed,

				const Self = @This();

				fn unpack(self: Self) []u8 {
					std.debug.assert(self.len >= 0);

					return self.ptr[0 .. @intCast(self.len)];
				}
			},

			dynamic: struct {
				ptr: [*]io.Byte,
				len: u32,

				const Self = @This();

				fn typeinfo(self: Self) *const Typeinfo {
					return @as(**const Typeinfo, @ptrCast(@alignCast(self.ptr))).*;
				}

				fn unpack(self: Self) []io.Byte {
					return self.ptr[0 .. (@sizeOf(usize) + self.len)];
				}

				fn userdata(self: Self) []io.Byte {
					const unpacked = self.unpack();
					const address_size = @sizeOf(usize);

					std.debug.assert(unpacked.len >= address_size);

					return unpacked[address_size ..];
				}
			},
		},

		fn acquire(self: *Internal) *Object {
			self.ref_count += 1;

			return @ptrCast(self);
		}
	};

	fn allocate(allocator: std.mem.Allocator, data: Internal) std.mem.Allocator.Error!*Object {
		const copy = try allocator.create(Internal);

		errdefer allocator.destroy(copy);

		copy.* = data;

		return @ptrCast(copy);
	}

	fn internal(self: *const Object) *Internal {
		return @constCast(@ptrCast(@alignCast(self)));
	}

	///
	/// Returns `true` if the value of `self` is equivalent to `other`, otherwise `false`.
	///
	pub fn equals(self: *Object, other: *Object) bool {
		return switch (self.internal().payload) {
			.false => other.internal().payload == .false,
			.true => other.internal().payload == .true,

			.fixed => |lhs_fixed| switch (other.internal().payload) {
				.fixed => |rhs_fixed| rhs_fixed == lhs_fixed,
				.float => |rhs_float| rhs_float == @as(Float, @floatFromInt(lhs_fixed)),
				else => false,
			},

			.float => |lhs_float| switch (other.internal().payload) {
				.float => |rhs_float| rhs_float == lhs_float,
				.fixed => |rhs_fixed| @as(Float, @floatFromInt(rhs_fixed)) == lhs_float,
				else => false,
			},

			.symbol => |lhs_symbol| switch (other.internal().payload) {
				.symbol => |rhs_symbol| lhs_symbol == rhs_symbol,
				else => false,
			},

			.boxed => |boxed| if (boxed) |object| self.equals(object) else false,

			.vector2 => |lhs_vector2| switch (other.internal().payload) {
				.vector2 => |rhs_vector2| lhs_vector2[0] == rhs_vector2[0] and lhs_vector2[1] == rhs_vector2[1],
				else => false,
			},

			.vector3 => |lhs_vector3| switch (other.internal().payload) {
				.vector3 => |rhs_vector3| lhs_vector3[0] == rhs_vector3[0] and lhs_vector3[1] == rhs_vector3[1] and lhs_vector3[2] == rhs_vector3[2],
				else => false,
			},

			.syscall => |lhs_syscall| switch (other.internal().payload) {
				.syscall => |rhs_syscall| lhs_syscall == rhs_syscall,
				else => false,
			},

			.string => |lhs_string| switch (other.internal().payload) {
				.string => |rhs_string| io.are_equal(lhs_string.unpack(), rhs_string.unpack()),
				else => false,
			},

			.dynamic => |lhs_dynamic| switch (other.internal().payload) {
				.dynamic => |rhs_dynamic|
					lhs_dynamic.typeinfo() == rhs_dynamic.typeinfo() and
					lhs_dynamic.userdata().ptr == rhs_dynamic.userdata().ptr,

				else => false,
			},
		};
	}

	///
	/// Computes a hash for `self` based on the held type and value.
	///
	pub fn hash(self: *Object) usize {
		return switch (self.internal().payload) {
			.false => 1237,
			.true => 1231,
			.float => |float| @bitCast(float),
			.fixed => |fixed| @intCast(@as(u32, @bitCast(fixed))),
			.symbol => |symbol| @intFromPtr(symbol),
			.vector2 => |vector| @bitCast(vector),
			.vector3 => |vector| hashes.jenkins(@typeInfo(usize).Int, io.bytes_of(&vector)),
			.syscall => |syscall| @intFromPtr(syscall),
			.boxed => |boxed| @intFromPtr(boxed),
			.string => |string| hashes.djb2(@typeInfo(usize).Int, string.unpack()),
			.dynamic => |dynamic| @intFromPtr(dynamic.typeinfo()) ^ @intFromPtr(dynamic.userdata().ptr),
		};
	}

	///
	/// Returns the name of the type held in `self`.
	///
	pub fn typename(self: *Object) []const u8 {
		return switch (self.internal().payload) {
			.false => "false",
			.true => "true",
			.float => "float",
			.fixed => "fixed",
			.symbol => "symbol",
			.vector2 => "vector2",
			.vector3 => "vector3",
			.syscall => "syscall",
			.boxed => "boxed",
			.string => "string",
			.dynamic => "dynamic",
		};
	}

	///
	/// Checks if `self` is a dynamic object with type information matching `typeinfo`, returning the userdata buffer if
	/// it is or `null` if it is not.
	///
	pub fn is_dynamic(self: *Object, comptime Type: type) ?*Type {
		return switch (self.internal().payload) {
			.dynamic => |dynamic| @as(?*Type, if (dynamic.typeinfo() == &Type.typeinfo) @ptrCast(@alignCast(dynamic.userdata())) else null),
			else => null,
		};
	}

	///
	/// Checks if `self` is a false object (i.e. any object that evaluates to false), returning `true` if it is or
	/// `false` if it is not.
	///
	pub fn is_false(self: *Object) bool {
		return !self.is_true();
	}

	///
	/// Checks if `self` is a fixed number object, returning the value if it is or `null` if it is not.
	///
	pub fn is_fixed(self: *Object) ?Fixed {
		return switch (self.internal().payload) {
			.fixed => |fixed| fixed,
			else => null,
		};
	}

	///
	/// Checks if `self` is a numeric object (i.e. any object that behaves like a number), returning the supertype for
	/// further inspection if it is or `null` if it is not.
	///
	pub fn is_numeric(self: *Object) ?Numeric {
		return switch (self.internal().payload) {
			.fixed => |fixed| .{.fixed = fixed},
			.float => |float| .{.float = float},
			.vector2 => |vector2| .{.vector2 = vector2},
			.vector3 => |vector3| .{.vector3 = vector3},
			else => null,
		};
	}

	///
	/// Checks if `self` is a string text object, returning the value if it is or `null` if it is not.
	///
	pub fn is_string(self: *Object) ?[]const u8 {
		return switch (self.internal().payload) {
			.string => |string| get: {
				std.debug.assert(string.len > -1);

				break: get string.ptr[0 .. @intCast(string.len)];
			},

			.symbol => |symbol| io.slice_sentineled(@as(u8, 0), symbol),
			else => null,
		};
	}

	///
	/// Checks if `self` is a symbo text object, returning the value if it is or `null` if it is not.
	///
	pub fn is_symbol(self: *Object) ?[*:0]const u8 {
		return switch (self.internal().payload) {
			.symbol => |symbol| symbol,
			else => null,
		};
	}

	///
	/// Checks if `self` is a true object (i.e. any object that evaluates to true), returning `true` if it is and
	/// `false` if it is not.
	///
	pub fn is_true(self: *Object) bool {
		return switch (self.internal().payload) {
			.false => false,
			.true => true,
			.float => |float| float != 0,
			.fixed => |fixed| fixed != 0,
			.symbol => true,
			.vector2 => |vector| vector[0] != 0 or vector[1] != 0,
			.vector3 => |vector| vector[0] != 0 or vector[1] != 0 or vector[2] != 0,
			.syscall => true,
			.boxed => |boxed| boxed != null,
			.string => |string| string.len != 0,
			.dynamic => true,
		};
	}

	///
	/// Checks if `self` is a 2-component vector number object, returning the value if it is or `null` if it is not.
	///
	pub fn is_vector2(self: *Object) ?Vector2 {
		return switch (self.internal().payload) {
			.vector2 => |vector2| vector2,
			else => null,
		};
	}

	///
	/// Checks if `self` is a 3-component vector number object, returning the value if it is or `null` if it is not.
	///
	pub fn is_vector3(self: *Object) ?Vector3 {
		return switch (self.internal().payload) {
			.vector3 => |vector3| vector3,
			else => null,
		};
	}
};

///
/// Runtime environment virtual machine state.
///
/// The environment operates on a mostly stack-based architecture, with nearly all functions "returning" computations by
/// writing them to the local values stack within. Functions may be chained together to perform sequential mutations on
/// data and either retrieved or discarded once the computation is done via [Runtime.pop] or [Runtime.discard]
/// respectively.
///
/// *Note* that improperly cleaning up data pushed to the stack can result in accidental stack overflows, however also
/// note that all locals are always cleaned when exiting their call frame.
///
pub const Runtime = struct {
	allocator: std.mem.Allocator,
	options: Options,
	interned_symbols: map.StringTable([:0]u8),
	locals: LocalList,
	frames: FrameList,

	const FrameList = stack.Sequential(struct {
		callable: *Object,
		locals_top: usize,
		arg_count: u8,

		const Self = @This();

		fn args_of(self: Self, env: *Runtime) []?*Object {
			return env.locals.values[self.locals_top .. (self.locals_top + self.arg_count)];
		}

		fn locals_of(self: Self, env: *Runtime) []?*Object {
			return env.locals.values[self.locals_top ..];
		}
	});

	const LocalList = stack.Sequential(?*Object);

	///
	/// Optional settings for a [Runtime].
	///
	pub const Options = struct {
		file_system: file.System = .data,
	};

	///
	/// Attempts to push the argument located at `arg_index` to the top of `self`, pushing `null` instead if the given
	/// argument does not exist or was provided as a nil value.
	///
	/// Arguments are indexed according to the order they are passed, with `0` referring to the first argument position.
	///
	/// A [Error] is returned if `self` is out of memory or the virtual machine is not inside a managed call
	/// frame.
	///
	/// `self` is returned for function chaining.
	///
	pub fn arg_get(self: *Runtime, arg_index: Local) Error!*Runtime {
		const frame = self.frames.peek() orelse {
			return self.raise(error.IllegalState, "cannot get args outside of a call frame", .{});
		};

		const args = frame.args_of(self);

		if (arg_index < args.len) {
			return self.push(args[arg_index]);
		}

		try self.locals.push(null);

		return self;
	}

	///
	/// Attempts to pop the top-most value of `self` and call it with `local_arg_count` as the number of locals prior to
	/// it in `self` that are intended to be arguments to it. Once the callable returns, the locals marked as arguments
	/// are popped as well.
	///
	/// A `local_arg_count` of `0` will call the function with no arguments and pop nothing other than the callable from
	/// `self` during invocation.
	///
	/// A [Error] is returned if `self` is out of memory, the top-most local is nil or not callable, or the
	/// callable raises a runtime error during invocation.
	///
	/// `self` is returned for function chaining.
	///
	pub fn call(self: *Runtime, local_arg_count: Local) Error!*Runtime {
		const callable = try self.expect_object(self.pop());

		defer self.release(callable);

		const result = get_result: {
			try self.frames.push(.{
				.locals_top = self.locals.values.len - local_arg_count,
				.callable = callable.internal().acquire(),
				.arg_count = local_arg_count,
			});

			defer {
				const popped_frame = self.frames.pop().?;

				self.release(popped_frame.callable);
				std.debug.assert(popped_frame.locals_top <= self.locals.values.len);

				var to_discard = self.locals.values.len - popped_frame.locals_top;

				while (to_discard != 0) : (to_discard -= 1) {
					self.discard();
				}
			}

			const frame = &self.frames.values[self.frames.values.len - 1];

			break: get_result try switch (frame.callable.internal().payload) {
				.syscall => |syscall| syscall(self),

				.dynamic => |dynamic| dynamic.typeinfo().call(.{
					.userdata = dynamic.userdata(),
					.env = self,
				}),

				else => self.raise(error.TypeMismatch, "{typename} is not callable", .{
					.typename = frame.callable.typename(),
				}),
			};
		};

		errdefer {
			if (result) |object| {
				self.release(object);
			}
		}

		try self.locals.push(result);

		return self;
	}

	pub fn count(self: *Runtime, countable: *Object) Error!Fixed {
		return switch (countable.internal().payload) {
			.dynamic => |dynamic| dynamic.typeinfo().count(.{
				.userdata = dynamic.userdata(),
				.env = self,
			}),

			else => self.raise(error.TypeMismatch, "{typename} is not countable", .{
				.typename = countable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop and compare the top-most local in `self` with `rhs_comparable`, pushing a true value if the
	/// local is greater than `rhs_comparable`, otherwise pushing a false value.
	///
	/// A [Error] is returned if `self` is out of memory, the top-most local is nil or not comparable, or
	/// `rhs_comparable` is not comparable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn compare_greater(self: *Runtime, rhs_comparable: *Object) Error!*Runtime {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) > @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) > rhs_float),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float > rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float > @as(Float, @floatFromInt(rhs_fixed))),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			else => self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop and compare the top-most local in `self` with `rhs_comparable`, pushing a true value if the
	/// local is greater than or equal to `rhs_comparable`, otherwise pushing a false value.
	///
	/// A [Error] is returned if `self` is out of memory, the top-most local is nil or not comparable, or
	/// `rhs_comparable` is not comparable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn compare_greater_equals(self: *Runtime, rhs_comparable: *Object) Error!*Runtime {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) >= @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) >= rhs_float),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float >= rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float >= @as(Float, @floatFromInt(rhs_fixed))),

				else => self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			else => self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop and compare the top-most local in `self` with `rhs_comparable`, pushing a true value if the
	/// local is less than `rhs_comparable`, otherwise pushing a false value.
	///
	/// A [Error] is returned if `self` is out of memory, the top-most local is nil or not comparable, or
	/// `rhs_comparable` is not comparable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn compare_less(self: *Runtime, rhs_comparable: *Object) Error!*Runtime {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) < @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) < rhs_float),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float < rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float < @as(Float, @floatFromInt(rhs_fixed))),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			else => return self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop and compare the top-most local in `self` with `rhs_comparable`, pushing a true value if the
	/// local is less than or equal to `rhs_comparable`, otherwise pushing a false value.
	///
	/// A [Error] is returned if `self` is out of memory, the top-most local is nil or not comparable, or
	/// `rhs_comparable` is not comparable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn compare_less_equals(self: *Runtime, rhs_comparable: *Object) Error!*Runtime {
		const lhs_comparable = try self.expect_object(self.pop());

		defer self.release(lhs_comparable);

		return switch (lhs_comparable.internal().payload) {
			.fixed => |lhs_fixed| switch (rhs_comparable.internal().payload) {
				.fixed => |rhs_fixed| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) <= @as(Float, @floatFromInt(rhs_fixed))),
				.float => |rhs_float| self.new_boolean(@as(Float, @floatFromInt(lhs_fixed)) <= rhs_float),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			.float => |lhs_float| switch (rhs_comparable.internal().payload) {
				.float => |rhs_float| self.new_boolean(lhs_float <= rhs_float),
				.fixed => |rhs_fixed| self.new_boolean(lhs_float <= @as(Float, @floatFromInt(rhs_fixed))),

				else => return self.raise(error.TypeMismatch, "right-hand {typename} is not comparable", .{
					.typename = rhs_comparable.typename(),
				}),
			},

			else => return self.raise(error.TypeMismatch, "left-hand {typename} is not comparable", .{
				.typename = lhs_comparable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop `local_concat_count` number of locals from the stack and concatenate them together as one long
	/// string object, pushing the result to the top of `self`.
	///
	/// A `local_concat_count` of `0` will push an empty string object and pop nothing from `self`.
	///
	/// A [Error] is returned if `self` is out of memory or one of the concatenated objects raises an error from
	/// within it's to_string implementation.
	///
	/// `self` is returned for function chaining.
	///
	pub fn concat(self: *Runtime, local_concat_count: Local) Error!*Runtime {
		if (local_concat_count == 0) {
			return self.new_string("");
		}

		const concated_buffer = build_buffer: {
			var concat_buffer = stack.Sequential(u8){.allocator = self.allocator};

			errdefer concat_buffer.deinit();

			const concat_values = self.locals.values[(self.locals.values.len - local_concat_count) .. self.locals.values.len];

			for (concat_values) |concat_value| {
				const string = (try (try self.push(concat_value)).to_string()).pop().?;

				defer self.release(string);

				try concat_buffer.push_all(string.is_string().?);
			}

			{
				var concated_value_count = concat_values.len;

				while (concated_value_count != 0) : (concated_value_count -= 1) {
					self.discard();
				}
			}

			break: build_buffer try concat_buffer.to_allocation(concat_buffer.values.len, 0);
		};

		errdefer self.allocator.free(concated_buffer);

		const string = try Object.allocate(self.allocator, .{
			.ref_count = 1,

			.payload = .{
				.string = .{
					.ptr = concated_buffer.ptr,
					.len = @intCast(concated_buffer.len),
				},
			},
		});

		errdefer self.release(string);

		try self.locals.push(string);

		return self;
	}

	///
	/// Deinitializes `self`, freeing all allocated virtual machine resources.
	///
	pub fn deinit(self: *Runtime) void {
		while (self.locals.pop()) |local| {
			if (local.*) |ref| {
				self.release(ref);
			}
		}

		{
			var entries = self.interned_symbols.entries();

			while (entries.next()) |entry| {
				self.allocator.free(entry.value);
			}
		}

		self.interned_symbols.deinit();
		self.locals.deinit();
		self.frames.deinit();
	}

	///
	/// Pops the top-most value of `self` and releases it.
	///
	pub fn discard(self: *Runtime) void {
		if (self.pop()) |popped| {
			self.release(popped);
		}
	}

	pub fn equals_nil(self: *Runtime) Error!*Runtime {
		if (self.pop()) |lhs_object| {
			defer self.release(lhs_object);

			return self.new_boolean(false);
		}

		return self.new_boolean(true);
	}

	pub fn equals_object(self: *Runtime, rhs_object: *Object) Error!*Runtime {
		if (self.pop()) |lhs_object| {
			defer self.release(lhs_object);

			return self.new_boolean(lhs_object.equals(rhs_object));
		}

		return self.new_boolean(false);
	}

	pub fn expect_dynamic(self: *Runtime, value: *Object, comptime Type: type) Error!*Type {
		return value.is_dynamic(Type) orelse self.raise(error.TypeMismatch, "expected dynamic object, not {typename}", .{
			.typename = value.typename(),
		});
	}

	pub fn expect_float(self: *Runtime, value: *Object) Error!Float {
		return switch (value.internal().payload) {
			.fixed => |fixed| @floatFromInt(fixed),
			.float => |float| float,

			else => self.raise(error.TypeMismatch, "expected float type, not {typename}", .{
				.typename = value.typename(),
			}),
		};
	}

	pub fn expect_fixed(self: *Runtime, value: *Object) Error!Fixed {
		return value.is_fixed() orelse self.raise(error.TypeMismatch, "expected fixed type, not {typename}", .{
			.typename = value.typename(),
		});
	}

	pub fn expect_numeric(self: *Runtime, value: *Object) Error!Numeric {
		return value.is_numeric() orelse self.raise(error.TypeMismatch, "expected numeric type, not {typename}", .{
			.typename = value.typename(),
		});
	}

	pub fn expect_object(self: *Runtime, value: ?*Object) Error!*Object {
		return value orelse self.raise(error.TypeMismatch, "expected object type, not nil", .{});
	}

	pub fn expect_string(self: *Runtime, value: *Object) Error![]const u8 {
		return value.is_string() orelse self.raise(error.TypeMismatch, "expected string type, not {typename}", .{
			.typename = value.typename(),
		});
	}

	pub fn expect_symbol(self: *Runtime, value: *Object) Error![*:0]const u8 {
		return value.is_symbol() orelse self.raise(error.TypeMismatch, "expected symbol type, not {typename}", .{
			.typename = value.typename(),
		});
	}

	///
	/// Attempts to pop the top-most value in `self` and add `rhs_fixed` to it, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_fixed` is not addable with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn fixed_add(self: *Runtime, rhs_fixed: Fixed) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (scalars.add(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) + @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2 + @as(Vector2, @splat(@floatFromInt(rhs_fixed)))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3 + @as(Vector3, @splat(@floatFromInt(rhs_fixed)))),

			else => self.raise(error.TypeMismatch, "fixed types are not addable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and divide it with `rhs_fixed` to it, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_fixed` is not divisible with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn fixed_divide(self: *Runtime, rhs_fixed: Fixed) Error!*Runtime {
		if (rhs_fixed == 0) {
			return self.raise(error.TypeMismatch, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) / @as(Float, @floatFromInt(rhs_fixed))),
			.float => |lhs_float| self.new_float(lhs_float / @as(Float, @floatFromInt(rhs_fixed))),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2 / @as(Vector2, @splat(@floatFromInt(rhs_fixed)))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3 / @as(Vector3, @splat(@floatFromInt(rhs_fixed)))),

			else => self.raise(error.TypeMismatch, "fixed types are not divisible with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and multiply it with `rhs_fixed` to it, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_fixed` is not multiplicable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn fixed_multiply(self: *Runtime, rhs_fixed: Fixed) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (scalars.mul(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) * @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2 * @as(Vector2, @splat(@floatFromInt(rhs_fixed)))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3 * @as(Vector3, @splat(@floatFromInt(rhs_fixed)))),

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and subtract it with `rhs_fixed`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_fixed` is not subtractable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn fixed_subtract(self: *Runtime, rhs_fixed: Fixed) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| if (scalars.sub(lhs_fixed, rhs_fixed)) |result|
					self.new_fixed(result)
				else
					self.new_float(@as(Float, @floatFromInt(lhs_fixed)) - @as(Float, @floatFromInt(rhs_fixed))),

			.float => |lhs_float| self.new_float(lhs_float + @as(Float, @floatFromInt(rhs_fixed))),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2 - @as(Vector2, @splat(@floatFromInt(rhs_fixed)))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3 - @as(Vector3, @splat(@floatFromInt(rhs_fixed)))),

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and add it with `rhs_float`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_float` is not addable with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn float_add(self: *Runtime, rhs_float: Float) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) + rhs_float),
			.float => |lhs_float| self.new_float(lhs_float + lhs_float),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2 + @as(Vector2, @splat(@floatCast(rhs_float)))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3 + @as(Vector3, @splat(@floatCast(rhs_float)))),

			else => self.raise(error.TypeMismatch, "fixed types are not addable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and divide it with `rhs_float` to it, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_float` is not divisible with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn float_divide(self: *Runtime, rhs_float: Float) Error!*Runtime {
		if (rhs_float == 0) {
			return self.raise(error.TypeMismatch, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) / rhs_float),
			.float => |lhs_float| self.new_float(lhs_float / lhs_float),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2 / @as(Vector2, @splat(@floatCast(rhs_float)))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3 / @as(Vector3, @splat(@floatCast(rhs_float)))),

			else => self.raise(error.TypeMismatch, "fixed types are not divisible with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and multiply it with `rhs_float`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_float` is not multiplicable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn float_multiply(self: *Runtime, rhs_float: Float) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) * rhs_float),
			.float => |lhs_float| self.new_float(lhs_float * lhs_float),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2 * @as(Vector2, @splat(@floatCast(rhs_float)))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3 * @as(Vector3, @splat(@floatCast(rhs_float)))),

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and subtract it with `rhs_float`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_float` is not subtractable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn float_subtract(self: *Runtime, rhs_float: Float) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return switch (addable.internal().payload) {
			.fixed => |lhs_fixed| self.new_float(@as(Float, @floatFromInt(lhs_fixed)) - rhs_float),
			.float => |lhs_float| self.new_float(lhs_float - lhs_float),
			.vector2 => |lhs_vector2| self.new_vector2(lhs_vector2 - @as(Vector2, @splat(@floatCast(rhs_float)))),
			.vector3 => |lhs_vector3| self.new_vector3(lhs_vector3 - @as(Vector3, @splat(@floatCast(rhs_float)))),

			else => self.raise(error.TypeMismatch, "fixed types are not multiplicable with {typename}", .{
				.typename = addable.typename(),
			}),
		};
	}

	///
	/// TODO: doc comment.
	///
	/// A [Error] is returned if `self` is out of memory, the top-most local is nil or not indexable, or the
	/// object-specific indexing operation raised an error.
	///
	/// `self` is returned for function chaining.
	///
	pub fn index_get(self: *Runtime, indexable: *Object) Error!*Runtime {
		const vector_max = 3;
		const index = try self.expect_object(self.pop());

		defer self.release(index);

		switch (indexable.internal().payload) {
			.vector2 => |vector2| {
				const swizzle_symbol = try self.expect_symbol(index);
				var swizzle_buffer = [_]f32{0} ** vector_max;
				var swizzle_count = @as(usize, 0);

				while (true) : (swizzle_count += 1) {
					if (swizzle_count > swizzle_buffer.len) {
						try self.locals.push(null);

						return self;
					}

					swizzle_buffer[swizzle_count] = switch (swizzle_symbol[swizzle_count]) {
						'x' => vector2[0],
						'y' => vector2[1],

						0 => return switch (swizzle_count) {
							1 => self.new_float(swizzle_buffer[0]),
							2 => self.new_vector2(.{swizzle_buffer[0], swizzle_buffer[1]}),
							3 => self.new_vector3(swizzle_buffer),
							else => unreachable,
						},

						else => return self.raise(error.BadOperation, "no such vector2 index or swizzle: `{index}`", .{
							.index = io.slice_sentineled(@as(u8, 0), swizzle_symbol),
						}),
					};
				}
			},

			.vector3 => |vector3| {
				const swizzle_symbol = try self.expect_symbol(index);
				var swizzle_buffer = [_]f32{0} ** vector_max;
				var swizzle_count = @as(usize, 0);

				while (true) : (swizzle_count += 1) {
					if (swizzle_count > swizzle_buffer.len) {
						try self.locals.push(null);

						return self;
					}

					swizzle_buffer[swizzle_count] = switch (swizzle_symbol[swizzle_count]) {
						'x' => vector3[0],
						'y' => vector3[1],
						'z' => vector3[2],

						0 => return switch (swizzle_count) {
							1 => self.new_float(swizzle_buffer[0]),
							2 => self.new_vector2(.{swizzle_buffer[0], swizzle_buffer[1]}),
							3 => self.new_vector3(swizzle_buffer),
							else => unreachable,
						},

						else => return self.raise(error.BadOperation, "no such vector3 index or swizzle: `{index}`", .{
							.index = io.slice_sentineled(@as(u8, 0), swizzle_symbol),
						}),
					};
				}
			},

			.dynamic => |dynamic| {
				return self.push(get_value: {
					_ = try self.push(index);

					const value = try dynamic.typeinfo().get(.{
						.userdata = dynamic.userdata(),
						.env = self,
					});

					self.release(try self.expect_object(self.pop()));

					break: get_value value;
				});
			},

			else => return self.raise(error.TypeMismatch, "{typename} is not get-indexable", .{
				.typename = indexable.typename(),
			}),
		}
	}

	///
	/// TODO: doc comment.
	///
	/// A [Error] is returned if `self` is out of memory, the top-most local is nil or not indexable, or the
	/// object-specific indexing operation raised an error.
	///
	pub fn index_set(self: *Runtime, indexable: *Object, value: ?*Object) Error!void {
		switch (indexable.internal().payload) {
			.dynamic => |dynamic| {
				_ = try self.push(value);

				try dynamic.typeinfo().set(.{
					.userdata = dynamic.userdata(),
					.env = self,
				});

				if (self.pop()) |object| {
					self.release(object);
				}

				self.release(try self.expect_object(self.pop()));
			},

			else => return self.raise(error.TypeMismatch, "{typename} is not set-indexable", .{
				.typename = indexable.typename(),
			}),
		}
	}

	///
	/// Attempts to import a script from the import [file.Access] (specified during [Runtime.init]) from `file_system`
	/// and `file_path`.
	///
	/// TODO: comment externals.
	///
	/// If the loaded script is a plain text file, it will be read, parsed, and compiled into an optimized format before
	/// being executed.
	///
	/// If the loaded script is already compiled, it is immediately executed.
	///
	/// After completing execution, the return value is pushed to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory, no script at `script_url` could be found by the runtime, the
	/// script is a source file and contains invalid syntax, or the imported script raised an error during execution.
	///
	/// Any syntax errors are reported using [print_error] prior to raising a runtime error.
	///
	/// `self` is returned for function chaining.
	///
	pub fn import(self: *Runtime, script_url: file.URL, externals: []const Chunk.External) Error!*Runtime {
		{
			const callable = new_chunk: {
				var script_file = script_url.open() catch {
					return self.raise(error.BadOperation, "failed to open `{name}`", .{.name = script_url.path.data()});
				};

				defer script_file.close();

				const script_text = io.read_alloc(script_file.reader(), self.allocator) catch {
					return self.raise(error.BadOperation, "failed to read `{name}`", .{.name = script_url.path.data()});
				};

				defer self.allocator.free(script_text);

				var root = try tree.Root.init(self.allocator);

				defer root.deinit();

				{
					var stream = tokens.Stream{.source = script_text};

					root.parse(&stream) catch |parse_error| {
						for (root.error_messages.values) |error_message| {
							std.log.err("{s}", .{error_message});
						}

						return self.raise(parse_error, "failed to parse `{name}`", .{.name = script_url.path.data()});
					};
				}

				var chunk = try Chunk.init(self, script_url.path.data(), &root.environment, externals);

				errdefer chunk.deinit(self);

				break: new_chunk (try self.new_dynamic(chunk)).pop().?;
			};

			errdefer self.release(callable);

			try self.locals.push(callable);
		}

		return self.call(0);
	}

	///
	/// Attempts to initialize a new virtual machine instance with `allocator` as the allocation strategy, `frames_max`
	/// as the maximum call depth, and `options` as the optional values to pass to the VM state.
	///
	/// This function will fail if `allocator` cannot allocate the memory required to create the runtime environment.
	///
	pub fn init(allocator: std.mem.Allocator, frames_max: u32, options: Options) std.mem.Allocator.Error!Runtime {
		var frames = FrameList{.allocator = allocator};

		errdefer frames.deinit();

		try frames.grow(frames_max);

		var locals = LocalList{.allocator = allocator};

		errdefer locals.deinit();

		try locals.grow(local_max * frames_max);

		return .{
			.interned_symbols = .{
				.allocator = allocator,
				.traits = .{},
			},

			.locals = locals,
			.frames = frames,
			.allocator = allocator,
			.options = options,
		};
	}

	const Local = u8;

	const local_max = std.math.maxInt(Local);

	///
	/// Attempts to push the local located at `index` to the top of `self`.
	///
	/// Locals are indexed according to the order they are declared, with `0` referring to the first declared arg, let,
	/// or var in the current frame.
	///
	/// A [Error] is returned if `self` is out of memory, `index` does not refer to a valid local, or the
	/// virtual-machine is not in a managed call frame.
	///
	/// `self` is returned for function chaining.
	///
	pub fn local_get(self: *Runtime, index: Local) Error!*Runtime {
		const frame = self.frames.peek() orelse {
			return self.raise(error.IllegalState, "cannot get locals outside of a call frame", .{});
		};

		const locals = frame.locals_of(self);

		if (index >= locals.len) {
			return self.raise(error.IllegalState, "invalid local get", .{});
		}

		return self.push(locals[index]);
	}

	///
	/// Attempts to set the local located at `index` to `value`.
	///
	/// Locals are indexed according to the order they are declared, with `0` referring to the first declared arg, let,
	/// or var in the current frame.
	///
	/// A [Error] if`index` does not refer to a valid local, or the virtual-machine is not in a managed call
	/// frame.
	///
	/// `self` is returned for function chaining.
	///
	pub fn local_set(self: *Runtime, index: Local, value: ?*Object) Error!void {
		const frame = self.frames.peek() orelse {
			return self.raise(error.IllegalState, "cannot get locals outside of a call frame", .{});
		};

		const locals = frame.locals_of(self);

		if (index >= locals.len) {
			return self.raise(error.IllegalState, "invalid local get", .{});
		}

		const local = &locals[index];

		if (local.*) |previous_local| {
			self.release(previous_local);
		}

		local.* = if (value) |object| object.internal().acquire() else null;
	}

	///
	/// Attempts to push a copy of the top-most local in `self`.
	///
	/// A [Error] is returned if `self` is out of memory, the virtual-machine is not in a managed call frame, or
	/// there are no locals in the frame.
	///
	/// `self` is returned for function chaining.
	///
	pub fn local_top(self: *Runtime) Error!*Runtime {
		const frame = self.frames.peek() orelse {
			return self.raise(error.IllegalState, "cannot get locals outside of a call frame", .{});
		};

		const locals = frame.locals_of(self);

		if (locals.len != 0) {
			return self.raise(error.IllegalState, "invalid local top", .{});
		}

		return self.push(locals[locals.len - 1]);
	}

	///
	/// Pops the top-most local from `self`, negates it, and pushes the result.
	///
	/// A [Error] is returned if `self` is out of memory or the top-most local is nil or not negatable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn neg(self: *Runtime) Error!*Runtime {
		const negatable = try self.expect_object(self.pop());

		defer self.release(negatable);

		return switch (negatable.internal().payload) {
			.fixed => |fixed| self.new_fixed(-fixed),
			.float => |float| self.new_float(-float),

			else => self.raise(error.TypeMismatch, "{typename} is not scalar negatable", .{
				.typename = negatable.typename(),
			}),
		};
	}

	///
	/// Attempts to create a new true or false object according to `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_boolean(self: *Runtime, value: bool) Error!*Runtime {
		const boolean = try Object.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = if (value) .true else .false,
		});

		errdefer self.release(boolean);

		try self.locals.push(boolean);

		return self;
	}

	///
	/// Attempts to create a new boxed object holding a reference to `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_boxed(self: *Runtime, value: ?*Object) Error!*Runtime {
		const boxed = try Object.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.boxed = if (value) |object| object.internal().acquire() else null},
		});

		errdefer self.release(boxed);

		try self.locals.push(boxed);

		return self;
	}

	///
	/// Attempts to create a new dynamic object from `userdata` and `typeinfo`, pushing it to the top of `self`.
	///
	/// *Note* the size of the type specified in `userdata` must match the size described in `typeinfo` exactly.
	///
	/// A [RuntimeError] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_dynamic(self: *Runtime, userdata: anytype) Error!*Runtime {
		const Userdata = @TypeOf(userdata);
		const userdata_size = @sizeOf(Userdata);

		if (userdata_size > std.math.maxInt(u32)) {
			@compileError(@typeName(Userdata) ++ " is too big to be a dynamic object");
		}

		if (!@hasDecl(Userdata, "typeinfo") or @TypeOf(Userdata.typeinfo) != Typeinfo) {
			@compileError(@typeName(Userdata) ++ " must contain a `typeinfo` declaration of type `Typeinfo`");
		}

		const dynamic = new: {
			const dynamic = try self.allocator.alloc(io.Byte, @sizeOf(usize) + userdata_size);

			errdefer self.allocator.free(dynamic);

			@memcpy(dynamic, io.bytes_of(&&Userdata.typeinfo));
			@memcpy(dynamic[@sizeOf(usize) ..], io.bytes_of(&userdata));

			break: new try Object.allocate(self.allocator, .{
				.ref_count = 1,

				.payload = .{
					.dynamic = .{
						.ptr = dynamic.ptr,
						.len = userdata_size,
					},
				},
			});
		};

		errdefer self.release(dynamic);

		try self.locals.push(dynamic);

		return self;
	}

	///
	/// Attempts to create a new fixed-size number object from `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_fixed(self: *Runtime, value: Fixed) Error!*Runtime {
		const syscall = try Object.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.fixed = value},
		});

		errdefer self.release(syscall);

		try self.locals.push(syscall);

		return self;
	}

	///
	/// Attempts to create a new floating point number object from `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_float(self: *Runtime, value: Float) Error!*Runtime {
		const syscall = try Object.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.float = value},
		});

		errdefer self.release(syscall);

		try self.locals.push(syscall);

		return self;
	}

	///
	/// Attempts to create a new byte string object from `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_string(self: *Runtime, value: []const u8) Error!*Runtime {
		if (value.len > std.math.maxInt(Fixed)) {
			return error.OutOfMemory;
		}

		const string = new: {
			const string = try self.allocator.dupe(u8, value);

			errdefer self.allocator.free(string);

			break: new try Object.allocate(self.allocator, .{
				.payload = .{
					.string = .{
						.ptr = string.ptr,
						.len = @intCast(string.len),
					},
				},

				.ref_count = 1,
			});
		};

		errdefer self.release(string);

		try self.locals.push(string);

		return self;
	}

	///
	/// Attempts to create a new system-native function object from `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_syscall(self: *Runtime, value: *const Syscall) Error!*Runtime {
		const syscall = try Object.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.syscall = value},
		});

		errdefer self.release(syscall);

		try self.locals.push(syscall);

		return self;
	}

	///
	/// Attempts to create a new empty table object, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_table(self: *Runtime) Error!*Runtime {
		var table = Table.init(self);

		errdefer table.deinit(self);

		return self.new_dynamic(table);
	}

	///
	/// Attempts to create a new 2-component vector object from `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_vector2(self: *Runtime, value: Vector2) Error!*Runtime {
		const vector2 = try Object.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.vector2 = value},
		});

		errdefer self.release(vector2);

		try self.locals.push(vector2);

		return self;
	}

	///
	/// Attempts to create a new 3-component vector object from `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_vector3(self: *Runtime, value: Vector3) Error!*Runtime {
		const vector3 = try Object.allocate(self.allocator, .{
			.ref_count = 1,
			.payload = .{.vector3 = value},
		});

		errdefer self.release(vector3);

		try self.locals.push(vector3);

		return self;
	}

	///
	/// Attempts to create a new symbol object from `value`, pushing it to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn new_symbol(self: *Runtime, value: []const u8) Error!*Runtime {
		const symbol = try Object.allocate(self.allocator, .{
			.payload = .{
				.symbol = self.interned_symbols.lookup(value) orelse create: {
					const symbol_string = try self.allocator.dupeZ(u8, value);

					errdefer self.allocator.free(symbol_string);

					std.debug.assert(try self.interned_symbols.insert(symbol_string[0 .. value.len], symbol_string));

					break: create symbol_string;
				},
			},

			.ref_count = 1,
		});

		errdefer self.release(symbol);

		try self.locals.push(symbol);

		return self;
	}

	///
	/// Pops the top-most value of `self`, returning the value.
	///
	/// *Note* any returned non-null pointer must be released via [Runtime.release] or else the resources belonging
	/// to the objects will never be freed by the runtime.
	///
	pub fn pop(self: *Runtime) ?*Object {
		return self.locals.pop().?.*;
	}

	///
	/// Attempts to push `value` to the top of `self`.
	///
	/// A [Error] is returned if `self` is out of memory.
	///
	/// `self` is returned for function chaining.
	///
	pub fn push(self: *Runtime, value: ?*Object) Error!*Runtime {
		if (value) |object| {
			const acquired = object.internal().acquire();

			errdefer self.release(acquired);

			try self.locals.push(acquired);
		} else {
			try self.locals.push(null);
		}

		return self;
	}

	///
	/// Creates stack-unwinding runtime error from `error_value` with `format` and `args` as the formatted error message
	/// presented with the type of error raised and a full stack-trace.
	///
	/// The `error_value` passed is returned so that the error may be raised and returned in a single line.
	///
	pub fn raise(self: *Runtime, error_value: Error, comptime format: []const u8, args: anytype) Error {
		const formatted_message = utf8.alloc_formatted(self.allocator, format, args) catch |alloc_error| {
			return alloc_error;
		};

		defer self.allocator.free(formatted_message);

		std.log.err("{s}", .{formatted_message});

		if (!self.frames.is_empty()) {
			std.log.err("stack trace:", .{});

			var remaining_frames = self.frames.values.len;

			while (remaining_frames != 0) {
				remaining_frames -= 1;

				const callable = self.frames.values[remaining_frames].callable;
				const name = (try (try self.push(callable)).to_string()).pop().?;

				defer self.release(name);

				if (callable.is_dynamic(Chunk)) |chunk| {
					const line = chunk.lines.values[chunk.cursor];

					const chunk_name = try utf8.alloc_formatted(self.allocator, "{name}@{line_number}", .{
						.name = get_name: {
							const string = name.is_string();

							std.debug.assert(string != null);

							break: get_name string.?;
						},

						.line_number = line.number,
					});

					defer self.allocator.free(chunk_name);

					std.log.err("{s}", .{chunk_name});
				} else {
					std.log.err("{s}", .{name.is_string().?});
				}
			}
		}

		return error_value;
	}

	///
	/// Releases `value` from `self`, decrementing the reference count (if applicable), and releasing all resources
	/// belonging to the object once the reference count reaches zero.
	///
	pub fn release(self: *Runtime, value: *Object) void {
		std.debug.assert(value.internal().ref_count != 0);

		value.internal().ref_count -= 1;

		if (value.internal().ref_count == 0) {
			switch (value.internal().payload) {
				.false, .true, .float, .fixed, .symbol, .vector2, .vector3, .syscall => {},

				.boxed => |*boxed| {
					if (boxed.*) |boxed_value| {
						self.release(boxed_value);
					}
				},

				.string => |string| {
					std.debug.assert(string.len >= 0);
					self.allocator.free(string.ptr[0 .. @intCast(string.len)]);
				},

				.dynamic => |dynamic| {
					if (dynamic.typeinfo().destruct) |destruct| {
						destruct(.{
							.userdata = dynamic.userdata(),
							.env = self,
						});
					}

					self.allocator.free(dynamic.unpack());
				},
			}

			self.allocator.destroy(value.internal());
		}
	}

	///
	/// Attempts to pop the top-most value of `self`, convert it to a string object representation, and push the result.
	///
	/// A [Error] is returned if `self` is out of memory or the to_string implementation of the top-most value of
	/// `self` raised an error.
	///
	/// `self` is returned for function chaining.
	///
	pub fn to_string(self: *Runtime) Error!*Runtime {
		const decimal_format = utf8.DecimalFormat.default;
		const value = try self.expect_object(self.pop());

		defer self.release(value);

		return switch (value.internal().payload) {
			.false => self.new_string("false"),
			.true => self.new_string("true"),

			.fixed => |fixed| convert: {
				var string = io.FixedBuffer(32, @as(u8, 0)){};

				debug.assert_ok(decimal_format.print(string.writer(), fixed));

				break: convert self.new_string(string.values());
			},

			.float => |float| convert: {
				var string = io.FixedBuffer(32, @as(u8, 0)){};

				debug.assert_ok(decimal_format.print(string.writer(), float));

				break: convert self.new_string(string.values());
			},

			.boxed => |boxed| (try self.push(try self.expect_object(boxed))).to_string(),
			.symbol => |symbol| self.new_string(io.slice_sentineled(@as(u8, 0), symbol)),

			.string => acquire: {
				try self.locals.push(value.internal().acquire());

				break: acquire self;
			},

			.vector2 => |vector2| convert: {
				var string = io.FixedBuffer(64, @as(u8, 0)){};
				const x, const y = vector2;

				debug.assert_ok(utf8.print_formatted(string.writer(), "vec2({x}, {y})", .{
					.x = x,
					.y = y,
				}));

				break: convert self.new_string(string.values());
			},

			.vector3 => |vector3| convert: {
				var string = io.FixedBuffer(96, @as(u8, 0)){};
				const x, const y, const z = vector3;

				debug.assert_ok(utf8.print_formatted(string.writer(), "vec3({x}, {y}, {z})", .{
					.x = x,
					.y = y,
					.z = z,
				}));

				break: convert self.new_string(string.values());
			},

			.syscall => self.new_string("syscall"),

			.dynamic => |dynamic| dynamic_to_string: {
				const string = try dynamic.typeinfo().to_string(.{
					.userdata = dynamic.userdata(),
					.env = self,
				});

				errdefer self.release(string);

				try self.locals.push(string);

				break: dynamic_to_string self;
			},
		};
	}

	///
	/// Attempts to pop the top-most value in `self` and add it with `rhs_vector2`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_vector2` is not addable with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector2_add(self: *Runtime, rhs_vector2: Vector2) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2 + rhs_vector2)
			else
				self.raise(error.TypeMismatch, "vector2 types are not addable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and divide it with `rhs_vector2`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_vector2` is not divisible with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector2_divide(self: *Runtime, rhs_vector2: Vector2) Error!*Runtime {
		if (rhs_vector2[0] == 0 or rhs_vector2[1] == 0) {
			return self.raise(error.TypeMismatch, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2 / rhs_vector2)
			else
				self.raise(error.TypeMismatch, "vector2 types are not divisible with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and multiply it with `rhs_vector2`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_vector2` is not multiplicable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector2_multiply(self: *Runtime, rhs_vector2: Vector2) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2 * rhs_vector2)
			else
				self.raise(error.TypeMismatch, "vector2 types are not multiplicable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and subtract it with `rhs_vector2`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_vector2` is not subtractable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector2_subtract(self: *Runtime, rhs_vector2: Vector2) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector2()) |lhs_vector2|
				self.new_vector2(lhs_vector2 - rhs_vector2)
			else
				self.raise(error.TypeMismatch, "vector2 types are not multiplicable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and add it with `rhs_vector3`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_vector3` is not addable with the top-most local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector3_add(self: *Runtime, rhs_vector3: Vector3) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3 + rhs_vector3)
			else
				self.raise(error.TypeMismatch, "vector3 types are not addable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and divide it with `rhs_vector3`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_vector3` is not divisible with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector3_divide(self: *Runtime, rhs_vector3: Vector3) Error!*Runtime {
		if (rhs_vector3[0] == 0 or rhs_vector3[1] == 0 or rhs_vector3[2] == 0) {
			return self.raise(error.TypeMismatch, "cannot divide by zero", .{});
		}

		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3 / rhs_vector3)
			else
				self.raise(error.TypeMismatch, "vector3 types are not divisible with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and multiply it with `rhs_vector3`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_vector3` is not multiplicable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector3_multiply(self: *Runtime, rhs_vector3: Vector3) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3 * rhs_vector3)
			else
				self.raise(error.TypeMismatch, "vector3 types are not multiplicable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self` and subtract it with `rhs_vector3`, pushing the result.
	///
	/// A [Error] is returned if `self` is out of memory or `rhs_vector3` is not subtractable with the top-most
	/// local.
	///
	/// `self` is returned for function chaining.
	///
	pub fn vector3_subtract(self: *Runtime, rhs_vector3: Vector3) Error!*Runtime {
		const addable = try self.expect_object(self.pop());

		defer self.release(addable);

		return if (addable.is_vector3()) |lhs_vector3|
				self.new_vector3(lhs_vector3 - rhs_vector3)
			else
				self.raise(error.TypeMismatch, "vector3 types are not multiplicable with {typename}", .{
					.typename = addable.typename(),
				});
	}

	///
	/// Attempts to pop the top-most value in `self`, pushing the unboxed value.
	///
	/// A [Error] is returned if `self` is out of memory or the top-most value is nil or not unboxable.
	///
	/// `self` is returned for function chaining.
	///
	pub fn boxed_get(self: *Runtime) Error!*Runtime {
		const unboxable = try self.expect_object(self.pop());

		defer self.release(unboxable);

		return switch (unboxable.internal().payload) {
			.boxed => |boxed| self.push(boxed),

			else => self.raise(error.TypeMismatch, "{typename} is not unboxable", .{
				.typename = unboxable.typename(),
			}),
		};
	}

	///
	/// Attempts to pop the top-most value in `self`, unboxing it, and replacing the contents with `value`.
	///
	/// A [Error] is returned if the top-most value is nil or not unboxable.
	///
	pub fn boxed_set(self: *Runtime, value: ?*Object) Error!void {
		const unboxable = try self.expect_object(self.pop());

		defer self.release(unboxable);

		switch (unboxable.internal().payload) {
			.boxed => |*unboxed_value| {
				if (unboxed_value.*) |unboxed_object| {
					self.release(unboxed_object);
				}

				unboxed_value.* = if (value) |object| object.internal().acquire() else null;
			},

			else => return self.raise(error.TypeMismatch, "{typename} is not unboxable", .{
				.typename = unboxable.typename(),
			}),
		}
	}
};

///
/// Bridge function type between the virtual machine and native code.
///
pub const Syscall = fn (env: *Runtime) Error!?*Object;

///
/// Type information for dynamically created object types
///
pub const Typeinfo = struct {
	name: []const u8,
	destruct: ?*const fn (context: DestructContext) void = null,
	to_string: *const fn (context: ToStringContext) Error!*Object = default_to_string,
	count: *const fn (context: CountContext) Error!Fixed = default_count,
	call: *const fn (context: CallContext) Error!?*Object = default_call,
	get: *const fn (context: GetContext) Error!?*Object = default_get,
	set: *const fn (context: SetContext) Error!void = default_set,

	///
	/// Context used for calling objects.
	///
	pub const CallContext = struct {
		env: *Runtime,
		userdata: []io.Byte,
	};

	///
	///
	///
	pub const CountContext = struct {
		env: *Runtime,
		userdata: []io.Byte,
	};

	///
	/// Context used with objects that have special destruction semantics
	///
	pub const DestructContext = struct {
		env: *Runtime,
		userdata: []io.Byte,
	};

	///
	/// Context used for getting indices of objects.
	///
	pub const GetContext = struct {
		env: *Runtime,
		userdata: []io.Byte,

		///
		/// Pushes the index object of the get operation to the top of the stack.
		///
		/// *Note* the index is guaranteed to be a non-`null` value.
		///
		/// A [Error] is returned if the virtual machine is out of memory.
		///
		/// The [Runtime] is returned for function chaining.
		///
		pub fn push_index(self: *const GetContext) Error!*Runtime {
			std.debug.assert(self.env.locals.values.len > 0);

			return self.env.push(self.env.locals.values[self.env.locals.values.len - 1]);
		}
	};

	///
	/// Context used for setting indices of objects to new values.
	///
	pub const SetContext = struct {
		env: *Runtime,
		userdata: []io.Byte,

		///
		/// Pushes the index object of the set operation to the top of the stack.
		///
		/// *Note* the index is guaranteed to be a non-`null` value.
		///
		/// A [Error] is returned if the virtual machine is out of memory.
		///
		/// The [Runtime] is returned for function chaining.
		///
		pub fn push_index(self: *const SetContext) Error!*Runtime {
			std.debug.assert(self.env.locals.values.len > 0);

			return self.env.push(self.env.locals.values[self.env.locals.values.len - 2]);
		}

		///
		/// Pushes the value of the set operation to the top of the stack.
		///
		/// A [Error] is returned if the virtual machine is out of memory.
		///
		/// The [Runtime] is returned for function chaining.
		///
		pub fn push_value(self: *const SetContext) Error!*Runtime {
			std.debug.assert(self.env.locals.values.len > 1);

			return self.env.push(self.env.locals.values[self.env.locals.values.len - 1]);
		}
	};

	///
	/// Context used for converting objects to string representations.
	///
	pub const ToStringContext = struct {
		env: *Runtime,
		userdata: []io.Byte,
	};

	fn default_call(context: CallContext) Error!?*Object {
		return context.env.raise(error.BadOperation, "this dynamic object is not callable", .{});
	}

	fn default_count(context: CountContext) Error!Fixed {
		return context.env.raise(error.BadOperation, "this dynamic object is not countable", .{});
	}

	fn default_get(context: GetContext) Error!?*Object {
		return context.env.raise(error.BadOperation, "this dynamic object is not get-indexable", .{});
	}

	fn default_set(context: SetContext) Error!void {
		return context.env.raise(error.BadOperation, "this dynamic object is not set-indexable", .{});
	}

	fn default_to_string(context: ToStringContext) Error!*Object {
		return (try context.env.new_string("{}")).pop().?;
	}
};

///
/// 2-component vector type.
///
pub const Vector2 = @Vector(2, f32);

///
/// 3-component vector type.
///
pub const Vector3 = @Vector(3, f32);

///
/// Higher-level wrapper for [Runtime.index_get] that makes it easier to index [Fixed] keys of objects.
///
pub fn get_at(env: *Runtime, indexable: *Object, index: Fixed) Error!?*Object {
	const index_number = (try env.new_fixed(index)).pop();

	defer env.discard(index_number);

	return (try (try env.push(indexable)).index_get(index_number)).pop();
}

///
/// Higher-level wrapper for [Runtime.index_get] that makes it easier to index field keys of objects.
///
pub fn get_field(env: *Runtime, indexable: *Object, field: []const u8) Error!?*Object {
	const field_symbol = (try env.new_symbol(field)).pop().?;

	defer env.release(field_symbol);

	return (try (try env.push(indexable)).index_get(field_symbol)).pop();
}

///
/// Higher-level wrapper for [Runtime.index_get] that makes it easier to index string keys of objects.
///
pub fn get_key(env: *Runtime, indexable: *Object, key: []const u8) Error!?*Object {
	const key_string = (try env.new_string(key)).pop();

	defer env.discard(key_string);

	return (try (try env.push(indexable)).index_get(key_string)).pop();
}
