const coral = @import("coral");

const file = @import("../file.zig");

const script = @import("../script.zig");

const std = @import("std");

const tokens = @import("./tokens.zig");

const tree = @import("./tree.zig");

name: *script.Object,
arity: u8,
opcodes: coral.Stack(Opcode),
lines: coral.Stack(tokens.Line),
cursor: usize,
constants: coral.Stack(*script.Object),
bindings: []?*script.Object,
externals: *script.Object,

const Compiler = struct {
	chunk: *Self,
	env: *script.Runtime,

	fn compile_argument(self: *const Compiler, environment: *const tree.Environment, initial_argument: ?*const tree.Expr) script.Error!u8 {
		// TODO: Exceeding 255 arguments will make the VM crash.
		var maybe_argument = initial_argument;
		var argument_count = @as(u8, 0);

		while (maybe_argument) |argument| {
			try self.compile_expression(environment, argument, null);

			maybe_argument = argument.next;
			argument_count += 1;
		}

		return argument_count;
	}

	fn compile_expression(self: *const Compiler, environment: *const tree.Environment, expression: *const tree.Expr, name: ?[]const u8) script.Error!void {
		const number_format = coral.utf8.DecimalFormat{
			.delimiter = "_",
			.positive_prefix = .none,
		};

		switch (expression.kind) {
			.nil_literal => try self.chunk.write(expression.line, .push_nil),
			.true_literal => try self.chunk.write(expression.line, .push_true),
			.false_literal => try self.chunk.write(expression.line, .push_false),

			.number_literal => |literal| {
				for (literal) |codepoint| {
					if (codepoint == '.') {
						return self.chunk.write(expression.line, .{
							.push_const = try self.declare_float(number_format.parse(literal, script.Float) orelse unreachable),
						});
					}
				}

				try self.chunk.write(expression.line, .{
					.push_const = try self.declare_fixed(number_format.parse(literal, script.Fixed) orelse unreachable),
				});
			},

			.string_literal => |literal| {
				try self.chunk.write(expression.line, .{.push_const = try self.declare_string(literal)});
			},

			.vector2 => |vector2| {
				try self.compile_expression(environment, vector2.x, null);
				try self.compile_expression(environment, vector2.y, null);
				try self.chunk.write(expression.line, .push_vector2);
			},

			.vector3 => |vector3| {
				try self.compile_expression(environment, vector3.x, null);
				try self.compile_expression(environment, vector3.y, null);
				try self.compile_expression(environment, vector3.z, null);
				try self.chunk.write(expression.line, .push_vector3);
			},

			.string_template => {
				var current_expression = expression.next orelse {
					return self.chunk.write(expression.line, .{.push_const = try self.declare_string("")});
				};

				var component_count = @as(u8, 0);

				while (true) {
					try self.compile_expression(environment, current_expression, null);

					component_count += 1;

					current_expression = current_expression.next orelse {
						return self.chunk.write(expression.line, .{.push_concat = component_count});
					};
				}
			},

			.symbol_literal => |literal| {
				try self.chunk.write(expression.line, .{.push_const = try self.declare_symbol(literal)});
			},

			.table => |table| {
				var entries = table.nodes();
				var num_entries = @as(u16, 0);

				while (entries.next()) |entry| {
					try self.compile_expression(environment, entry.key, null);
					try self.compile_expression(environment, entry.value, null);

					num_entries = coral.scalars.add(num_entries, 1) orelse {
						return self.env.raise(error.OutOfMemory, "too many initializer values", .{});
					};
				}

				try self.chunk.write(expression.line, .{.push_table = num_entries});
			},

			.lambda_construct => |lambda_construct| {
				var chunk = try Self.init(self.env, name orelse "<lambda>", lambda_construct.environment, &.{});

				errdefer chunk.deinit(self.env);

				if (lambda_construct.environment.capture_count == 0) {
					try self.chunk.write(expression.line, .{.push_const = try self.declare_chunk(chunk)});
				} else {
					const lambda_captures = lambda_construct.environment.get_captures();
					var index = lambda_captures.len;

					while (index != 0) {
						index -= 1;

						try self.chunk.write(expression.line, switch (lambda_captures[index]) {
							.declaration_index => |declaration_index| .{.push_local = declaration_index},
							.capture_index => |capture_index| .{.push_binding = capture_index},
						});
					}

					try self.chunk.write(expression.line, .{.push_const = try self.declare_chunk(chunk)});
					try self.chunk.write(expression.line, .{.bind = lambda_construct.environment.capture_count});
				}
			},

			.binary_op => |binary_op| {
				try self.compile_expression(environment, binary_op.lhs_operand, null);
				try self.compile_expression(environment, binary_op.rhs_operand, null);

				try self.chunk.write(expression.line, switch (binary_op.operation) {
					.addition => .add,
					.subtraction => .sub,
					.multiplication => .mul,
					.divsion => .div,
					.greater_equals_comparison => .cge,
					.greater_than_comparison => .cgt,
					.equals_comparison => .eql,
					.less_than_comparison => .clt,
					.less_equals_comparison => .cle,
				});
			},

			.unary_op => |unary_op| {
				try self.compile_expression(environment, unary_op.operand, null);

				try self.chunk.write(expression.line, switch (unary_op.operation) {
					.boolean_negation => .not,
					.numeric_negation => .neg,
				});
			},

			.invoke => |invoke| {
				const argument_count = try self.compile_argument(environment, invoke.argument);

				try self.compile_expression(environment, invoke.object, null);
				try self.chunk.write(expression.line, .{.call = argument_count});
			},

			.group => |group| try self.compile_expression(environment, group, null),

			.declaration_get => |declaration_get| {
				if (get_local_index(environment, declaration_get.declaration)) |index| {
					if (is_declaration_boxed(declaration_get.declaration)) {
						try self.chunk.write(expression.line, .{.push_local = index});
						try self.chunk.write(expression.line, .get_box);
					} else {
						try self.chunk.write(expression.line, .{.push_local = index});
					}

					return;
				}

				if (try get_binding_index(environment, declaration_get.declaration)) |index| {
					try self.chunk.write(expression.line, .{.push_binding = index});

					if (is_declaration_boxed(declaration_get.declaration)) {
						try self.chunk.write(expression.line, .get_box);
					}

					return;
				}

				return self.env.raise(error.IllegalState, "local out of scope", .{});
			},

			.declaration_set => |declaration_set| {
				if (get_local_index(environment, declaration_set.declaration)) |index| {
					if (is_declaration_boxed(declaration_set.declaration)) {
						try self.chunk.write(expression.line, .{.push_local = index});
						try self.compile_expression(environment, declaration_set.assign, null);
						try self.chunk.write(expression.line, .set_box);
					} else {
						try self.compile_expression(environment, declaration_set.assign, null);
						try self.chunk.write(expression.line, .{.set_local = index});
					}

					return;
				}

				if (try get_binding_index(environment, declaration_set.declaration)) |index| {
					try self.compile_expression(environment, declaration_set.assign, null);
					try self.chunk.write(expression.line, .{.push_binding = index});

					if (is_declaration_boxed(declaration_set.declaration)) {
						try self.chunk.write(expression.line, .set_box);
					}

					return;
				}

				return self.env.raise(error.IllegalState, "local out of scope", .{});
			},

			.field_get => |field_get| {
				try self.chunk.write(expression.line, .{.push_const = try self.declare_symbol(field_get.identifier)});
				try self.compile_expression(environment, field_get.object, null);
				try self.chunk.write(expression.line, .get_dynamic);
			},

			.field_set => |field_set| {
				try self.chunk.write(expression.line, .{.push_const = try self.declare_symbol(field_set.identifier)});
				try self.compile_expression(environment, field_set.assign, null);
				try self.compile_expression(environment, field_set.object, null);
				try self.chunk.write(expression.line, .set_dynamic);
			},

			.subscript_get => |subscript_get| {
				try self.compile_expression(environment, subscript_get.index, null);
				try self.compile_expression(environment, subscript_get.object, null);
				try self.chunk.write(expression.line, .get_dynamic);
			},

			.subscript_set => |subscript_set| {
				try self.compile_expression(environment, subscript_set.index, null);
				try self.compile_expression(environment, subscript_set.assign, null);
				try self.compile_expression(environment, subscript_set.object, null);
				try self.chunk.write(expression.line, .set_dynamic);
			},

			.external_get => |external_get| {
				try self.chunk.write(expression.line, .{.push_const = try self.declare_symbol(external_get.name)});
				try self.chunk.write(expression.line, .get_external);
			},
		}
	}

	pub fn compile_environment(self: *const Compiler, environment: *const tree.Environment) script.Error!void {
		if (environment.statement) |statement| {
			const last_statement = try self.compile_statement(environment, statement);

			if (last_statement.kind != .@"return") {
				try self.chunk.write(last_statement.line, .push_nil);
			}
		}
	}

	fn compile_statement(self: *const Compiler, environment: *const tree.Environment, initial_statement: *const tree.Stmt) script.Error!*const tree.Stmt {
		var current_statement = initial_statement;

		while (true) {
			switch (current_statement.kind) {
				.@"return" => |@"return"| {
					if (@"return".returned_expression) |expression| {
						try self.compile_expression(environment, expression, null);
					} else {
						try self.chunk.write(current_statement.line, .push_nil);
					}

					// TODO: Omit ret calls at ends of chunk.
					try self.chunk.write(current_statement.line, .ret);
				},

				.@"while" => |@"while"| {
					try self.compile_expression(environment, @"while".loop_expression, null);
					try self.chunk.write(current_statement.line, .{.jf = 0});

					const origin_index = @as(u16, @intCast(self.chunk.opcodes.values.len - 1));

					_ = try self.compile_statement(environment, @"while".loop);
					self.chunk.opcodes.values[origin_index].jf = @intCast(self.chunk.opcodes.values.len - 1);

					try self.compile_expression(environment, @"while".loop_expression, null);
					try self.chunk.write(current_statement.line, .{.jt = origin_index});
				},

				.@"if" => |@"if"| {
					try self.compile_expression(environment, @"if".then_expression, null);
					try self.chunk.write(current_statement.line, .{.jf = 0});

					const origin_index = @as(u16, @intCast(self.chunk.opcodes.values.len - 1));

					_ = try self.compile_statement(environment, @"if".@"then");
					self.chunk.opcodes.values[origin_index].jf = @intCast(self.chunk.opcodes.values.len - 1);

					if (@"if".@"else") |@"else"| {
						_ = try self.compile_statement(environment, @"else");
					}
				},

				.declare => |declare| {
					try self.compile_expression(environment, declare.initial_expression, declare.declaration.identifier);

					if (is_declaration_boxed(declare.declaration)) {
						try self.chunk.write(current_statement.line, .push_boxed);
					}
				},

				.top_expression => |top_expression| {
					try self.compile_expression(environment, top_expression, null);

					if (top_expression.kind == .invoke) {
						try self.chunk.write(current_statement.line, .pop);
					}
				},
			}

			current_statement = current_statement.next orelse return current_statement;
		}
	}

	const constants_max = @as(usize, std.math.maxInt(u16));

	fn declare_chunk(self: *const Compiler, chunk: Self) script.Error!u16 {
		if (self.chunk.constants.values.len == std.math.maxInt(u16)) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = @as(usize, std.math.maxInt(u16)),
			});
		}

		const constant = (try self.env.new_dynamic(chunk)).pop().?;

		errdefer self.env.release(constant);

		try self.chunk.constants.push(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_fixed(self: *const Compiler, fixed: script.Fixed) script.Error!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = (try self.env.new_fixed(fixed)).pop().?;

		errdefer self.env.release(constant);

		try self.chunk.constants.push(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_float(self: *const Compiler, float: script.Float) script.Error!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = (try self.env.new_float(float)).pop().?;

		errdefer self.env.release(constant);

		try self.chunk.constants.push(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_string(self: *const Compiler, string: []const u8) script.Error!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = (try self.env.new_string(string)).pop().?;

		errdefer self.env.release(constant);

		try self.chunk.constants.push(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_vector2(self: *const Compiler, vector: script.Vector2) script.Error!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = (try self.env.new_vector2(vector)).pop().?;

		errdefer self.env.release(constant);

		try self.chunk.constants.push(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_vector3(self: *const Compiler, vector: script.Vector3) script.Error!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = (try self.env.new_vector3(vector)).pop().?;

		errdefer self.env.release(constant);

		try self.chunk.constants.push(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn declare_symbol(self: *const Compiler, symbol: []const u8) script.Error!u16 {
		if (self.chunk.constants.values.len == constants_max) {
			return self.env.raise(error.BadSyntax, "chunks cannot contain more than {max} constants", .{
				.max = constants_max,
			});
		}

		const constant = (try self.env.new_symbol(symbol)).pop().?;

		errdefer self.env.release(constant);

		try self.chunk.constants.push(constant);

		return @intCast(self.chunk.constants.values.len - 1);
	}

	fn get_binding_index(environment: *const tree.Environment, declaration: *const tree.Declaration) script.Error!?u8 {
		var binding_index = @as(u8, 0);

		while (binding_index < environment.capture_count) : (binding_index += 1) {
			var capture = &environment.captures[binding_index];
			var target_environment = environment.enclosing orelse return null;

			while (capture.* == .capture_index) {
				capture = &target_environment.captures[capture.capture_index];
				target_environment = target_environment.enclosing orelse return null;
			}

			std.debug.assert(capture.* == .declaration_index);

			if (&target_environment.declarations[capture.declaration_index] == declaration) {
				return binding_index;
			}
		}

		return null;
	}

	fn get_local_index(environment: *const tree.Environment, declaration: *const tree.Declaration) ?u8 {
		var remaining = environment.declaration_count;

		while (remaining != 0) {
			remaining -= 1;

			if (&environment.declarations[remaining] == declaration) {
				return remaining;
			}
		}

		return null;
	}

	fn is_declaration_boxed(declaration: *const tree.Declaration) bool {
		return declaration.is.captured and !declaration.is.readonly;
	}
};

pub const Opcode = union (enum) {
	ret,
	pop,
	push_nil,
	push_true,
	push_false,
	push_const: u16,
	push_local: u8,
	push_top,
	push_vector2,
	push_vector3,
	push_table: u16,
	push_binding: u8,
	push_concat: u8,
	push_boxed,
	set_local: u8,
	get_dynamic,
	set_dynamic,
	get_external,
	get_box,
	set_box,
	call: u8,
	bind: u8,

	not,
	neg,

	add,
	sub,
	mul,
	div,

	eql,
	cgt,
	clt,
	cge,
	cle,

	jt: u16,
	jf: u16,
};

pub const External = struct {[]const u8, *script.Object};

const Self = @This();

pub fn deinit(self: *Self, env: *script.Runtime) void {
	while (self.constants.pop()) |constant| {
		env.release(constant.*);
	}

	self.constants.deinit();
	self.opcodes.deinit();
	self.lines.deinit();
	env.release(self.name);
	env.release(self.externals);

	if (self.bindings.len != 0) {
		for (self.bindings) |binding| {
			if (binding) |value| {
				env.release(value);
			}
		}

		env.allocator.free(self.bindings);
	}

	self.bindings = &.{};
}

pub fn dump(chunk: Self, env: *script.Runtime) script.Error!*script.Object {
	var opcode_cursor = @as(u32, 0);
	var buffer = coral.list.ByteStack.init(env.allocator);

	defer buffer.deinit();

	const writer = coral.list.stack_as_writer(&buffer);

	_ = coral.utf8.print_string(writer, "\n");

	while (opcode_cursor < chunk.opcodes.values.len) : (opcode_cursor += 1) {
		_ = coral.utf8.print_formatted(writer, "[{instruction}]: ", .{.instruction = opcode_cursor});

		_ = switch (chunk.opcodes.values[opcode_cursor]) {
			.ret => coral.utf8.print_string(writer, "ret\n"),
			.pop => coral.utf8.print_string(writer, "pop\n"),
			.push_nil => coral.utf8.print_string(writer, "push nil\n"),
			.push_true => coral.utf8.print_string(writer, "push true\n"),
			.push_false => coral.utf8.print_string(writer, "push false\n"),

			.push_const => |push_const| print: {
				const string_ref = (try (try env.push(try chunk.get_constant(env, push_const))).to_string()).pop().?;

				defer env.release(string_ref);

				const string = string_ref.is_string();

				break: print coral.utf8.print_formatted(writer, "push const ({value})\n", .{.value = string.?});
			},

			.push_local => |push_local| coral.utf8.print_formatted(writer, "push local ({local})\n", .{
				.local = push_local,
			}),

			.push_top => coral.utf8.print_string(writer, "push top\n"),

			.push_table => |push_table| coral.utf8.print_formatted(writer, "push table ({count})\n", .{
				.count = push_table,
			}),

			.push_boxed => coral.utf8.print_string(writer, "push boxed\n"),

			.push_binding => |push_binding| coral.utf8.print_formatted(writer, "push binding ({binding})\n", .{
				.binding = push_binding,
			}),

			.push_concat => |push_concat| coral.utf8.print_formatted(writer, "push concat ({count})\n", .{
				.count = push_concat,
			}),

			.push_builtin => |push_builtin| coral.utf8.print_formatted(writer, "push builtin ({builtin})\n", .{
				.builtin = switch (push_builtin) {
					.import => "import",
					.print => "print",
					.vec2 => "vec2",
					.vec3 => "vec3",
				},
			}),

			.bind => |bind| coral.utf8.print_formatted(writer, "bind ({count})\n", .{
				.count = bind,
			}),

			.set_local => |local_set| coral.utf8.print_formatted(writer, "set local ({local})\n", .{
				.local = local_set,
			}),

			.get_box => coral.utf8.print_string(writer, "get box\n"),
			.set_box => coral.utf8.print_string(writer, "set box\n"),
			.get_dynamic => coral.utf8.print_string(writer, "get dynamic\n"),
			.set_dynamic => coral.utf8.print_string(writer, "set dynamic\n"),
			.call => |call| coral.utf8.print_formatted(writer, "call ({count})\n", .{.count = call}),
			.not => coral.utf8.print_string(writer, "not\n"),
			.neg => coral.utf8.print_string(writer, "neg\n"),
			.add => coral.utf8.print_string(writer, "add\n"),
			.sub => coral.utf8.print_string(writer, "sub\n"),
			.mul => coral.utf8.print_string(writer, "mul\n"),
			.div => coral.utf8.print_string(writer, "div\n"),
			.eql => coral.utf8.print_string(writer, "eql\n"),
			.cgt => coral.utf8.print_string(writer, "cgt\n"),
			.clt => coral.utf8.print_string(writer, "clt\n"),
			.cge => coral.utf8.print_string(writer, "cge\n"),
			.cle => coral.utf8.print_string(writer, "cle\n"),
			.jf => |jf| coral.utf8.print_formatted(writer, "jf ({instruction})\n", .{.instruction = jf}),
			.jt => |jt| coral.utf8.print_formatted(writer, "jt ({instruction})\n", .{.instruction = jt}),
		};
	}

	return (try env.new_string(buffer.values)).pop().?;
}

pub fn execute(self: *Self, env: *script.Runtime) script.Error!?*script.Object {
	self.cursor = 0;

	while (self.cursor < self.opcodes.values.len) : (self.cursor += 1) {
		switch (self.opcodes.values[self.cursor]) {
			.ret => break,
			.pop => env.discard(),
			.push_nil => _ = try env.push(null),
			.push_true => _ = try env.new_boolean(true),
			.push_false => _ = try env.new_boolean(false),
			.push_const => |push_const| _ = try env.push(try self.get_constant(env, push_const)),
			.push_local => |push_local| _ = try env.local_get(push_local),
			.push_top => _ = try env.local_top(),

			.push_vector2 => {
				const y = try env.expect_float(try env.expect_object(env.pop()));
				const x = try env.expect_float(try env.expect_object(env.pop()));

				_ = try env.new_vector2(.{@floatCast(x), @floatCast(y)});
			},

			.push_vector3 => {
				const z = try env.expect_float(try env.expect_object(env.pop()));
				const y = try env.expect_float(try env.expect_object(env.pop()));
				const x = try env.expect_float(try env.expect_object(env.pop()));

				_ =  try env.new_vector3(.{@floatCast(x), @floatCast(y), @floatCast(z)});
			},

			.push_table => |push_table| {
				const table = (try env.new_table()).pop().?;

				defer env.release(table);

				var popped = @as(usize, 0);

				while (popped < push_table) : (popped += 1) {
					if (env.pop()) |object| {
						defer env.release(object);

						try env.index_set(table, object);
					} else {
						env.release(try env.expect_object(env.pop()));
					}
				}

				_ = try env.push(table);
			},

			.push_boxed => {
				const value = env.pop();

				defer {
					if (value) |object| {
						env.release(object);
					}
				}

				_ = try env.new_boxed(value);
			},

			.push_binding => |push_binding| _ = try env.push(try self.get_binding(env, push_binding)),
			.push_concat => |push_concat| _ = try env.concat(push_concat),

			.bind => |bind| {
				const callable = try env.expect_object(env.pop());

				defer env.release(callable);

				const chunk = try env.expect_dynamic(callable, Self);

				if (chunk.bindings.len != 0) {
					return env.raise(error.IllegalState, "cannot bind values to an already-bound chunk", .{});
				}

				chunk.bindings = try env.allocator.alloc(?*script.Object, bind);

				errdefer env.allocator.free(chunk.bindings);

				for (0 .. bind) |index| {
					const value = env.pop();

					errdefer {
						if (value) |object| {
							env.release(object);
						}
					}

					chunk.bindings[index] = value;
				}

				_ = try env.push(callable);
			},

			.set_local => |local_set| _ = try env.local_set(local_set, env.pop()),
			.get_box => _ = try env.boxed_get(),

			.set_box => {
				const value = env.pop();

				defer {
					if (value) |object| {
						env.release(object);
					}
				}

				try env.boxed_set(value);
			},

			.get_dynamic => {
				const indexable = try env.expect_object(env.pop());

				defer env.release(indexable);

				_ = try env.index_get(indexable);
			},

			.set_dynamic => {
				const indexable = try env.expect_object(env.pop());

				defer env.release(indexable);

				const value = env.pop();

				defer {
					if (value) |object| {
						env.release(object);
					}
				}

				try env.index_set(indexable, value);
			},

			.get_external => _ = try env.index_get(self.externals),
			.call => |call| _ = try env.call(call),

			.not => {
				const object = try env.expect_object(env.pop());

				defer env.release(object);

				_ = try env.new_boolean(object.is_false());
			},

			.neg => _ = try env.neg(),

			.add => {
				const addable = try env.expect_object(env.pop());

				defer env.release(addable);

				_ = switch (try env.expect_numeric(addable)) {
					.fixed => |fixed| try env.fixed_add(fixed),
					.float => |float| try env.float_add(float),
					.vector2 => |vector2| try env.vector2_add(vector2),
					.vector3 => |vector3| try env.vector3_add(vector3),
				};
			},

			.sub => {
				const subtractable = try env.expect_object(env.pop());

				defer env.release(subtractable);

				_ = switch (try env.expect_numeric(subtractable)) {
					.fixed => |fixed| try env.fixed_subtract(fixed),
					.float => |float| try env.float_subtract(float),
					.vector2 => |vector2| try env.vector2_subtract(vector2),
					.vector3 => |vector3| try env.vector3_subtract(vector3),
				};
			},

			.mul => {
				const multiplicable = try env.expect_object(env.pop());

				defer env.release(multiplicable);

				_ = switch (try env.expect_numeric(multiplicable)) {
					.fixed => |fixed| try env.fixed_multiply(fixed),
					.float => |float| try env.float_multiply(float),
					.vector2 => |vector2| try env.vector2_multiply(vector2),
					.vector3 => |vector3| try env.vector3_multiply(vector3),
				};
			},

			.div => {
				const divisible = try env.expect_object(env.pop());

				defer env.release(divisible);

				_ = switch (try env.expect_numeric(divisible)) {
					.fixed => |fixed| try env.fixed_divide(fixed),
					.float => |float| try env.float_divide(float),
					.vector2 => |vector2| try env.vector2_divide(vector2),
					.vector3 => |vector3| try env.vector3_divide(vector3),
				};
			},

			.eql => {
				if (env.pop()) |equatable| {
					defer env.release(equatable);

					_ = try env.equals_object(equatable);
				} else {
					_ = try env.equals_nil();
				}
			},

			.cgt => {
				const comparable = try env.expect_object(env.pop());

				defer env.release(comparable);

				_ = try env.compare_greater(comparable);
			},

			.clt => {
				const comparable = try env.expect_object(env.pop());

				defer env.release(comparable);

				_ = try env.compare_less(comparable);
			},

			.cge => {
				const comparable = try env.expect_object(env.pop());

				defer env.release(comparable);

				_ = try env.compare_greater_equals(comparable);
			},

			.cle => {
				const comparable = try env.expect_object(env.pop());

				defer env.release(comparable);

				_ = try env.compare_less_equals(comparable);
			},

			.jf => |jf| {
				if (env.pop()) |condition| {
					defer env.release(condition);

					if (condition.is_false()) {
						self.cursor = jf;
					}
				} else {
					self.cursor = jf;
				}
			},

			.jt => |jt| {
				if (env.pop()) |condition| {
					defer env.release(condition);

					if (condition.is_true()) {
						self.cursor = jt;
					}
				}
			},
		}
	}

	return env.pop();
}

fn get_binding(self: *Self, env: *script.Runtime, index: usize) script.Error!?*script.Object {
	if (index >= self.bindings.len) {
		return env.raise(error.IllegalState, "invalid binding", .{});
	}

	return self.bindings[index];
}

fn get_constant(self: *const Self, env: *script.Runtime, index: usize) script.Error!*script.Object {
	if (index >= self.constants.values.len) {
		return env.raise(error.IllegalState, "invalid constant", .{});
	}

	return self.constants.values[index];
}

pub fn init(env: *script.Runtime, name: []const u8, environment: *const tree.Environment, externals: []const External) script.Error!Self {
	const name_symbol = (try env.new_symbol(name)).pop().?;

	errdefer env.release(name_symbol);

	const externals_table = (try env.new_table()).pop().?;

	errdefer env.release(externals_table);

	for (0 .. externals.len) |i| {
		const external_name, const external_object = externals[i];

		try (try env.new_symbol(external_name)).index_set(externals_table, external_object);
	}

	var chunk = Self{
		.externals = externals_table,
		.name = name_symbol,
		.opcodes = .{.allocator = env.allocator},
		.constants = .{.allocator = env.allocator},
		.lines = .{.allocator = env.allocator},
		.bindings = &.{},
		.arity = environment.argument_count,
		.cursor = 0,
	};

	var compiler = Compiler{
		.chunk = &chunk,
		.env = env,
	};

	try compiler.compile_environment(environment);

	return chunk;
}

pub const typeinfo = script.Typeinfo{
	.name = "lambda",
	.destruct = typeinfo_destruct,
	.call = typeinfo_call,
	.to_string = typeinfo_to_string,
};

fn typeinfo_call(context: script.Typeinfo.CallContext) script.Error!?*script.Object {
	return @as(*Self, @ptrCast(@alignCast(context.userdata))).execute(context.env);
}

fn typeinfo_destruct(context: script.Typeinfo.DestructContext) void {
	@as(*Self, @ptrCast(@alignCast(context.userdata))).deinit(context.env);
}

fn typeinfo_to_string(context: script.Typeinfo.ToStringContext) script.Error!*script.Object {
	return (try (try context.env.push(@as(*Self, @ptrCast(@alignCast(context.userdata))).name)).to_string()).pop().?;
}

pub fn write(self: *Self, line: tokens.Line, opcode: Opcode) std.mem.Allocator.Error!void {
	try self.opcodes.push(opcode);
	try self.lines.push(line);
}
