const coral = @import("coral");

const std = @import("std");

pub const Line = struct {
	number: u32,
};

pub const Token = union(enum) {
	end,
	unknown: u8,
	newline,
	identifier: []const u8,
	builtin: []const u8,

	symbol_plus,
	symbol_minus,
	symbol_asterisk,
	symbol_forward_slash,
	symbol_paren_left,
	symbol_paren_right,
	symbol_bang,
	symbol_comma,
	symbol_at,
	symbol_brace_left,
	symbol_brace_right,
	symbol_bracket_left,
	symbol_bracket_right,
	symbol_period,
	symbol_colon,
	symbol_less_than,
	symbol_less_equals,
	symbol_greater_than,
	symbol_greater_equals,
	symbol_equals,
	symbol_double_equals,

	number: []const u8,
	string: []const u8,
	template_string: []const u8,

	keyword_nil,
	keyword_false,
	keyword_true,
	keyword_return,
	keyword_self,
	keyword_const,
	keyword_if,
	keyword_do,
	keyword_end,
	keyword_while,
	keyword_else,
	keyword_elif,
	keyword_var,
	keyword_vec2,
	keyword_vec3,
	keyword_let,
	keyword_lambda,

	pub fn text(self: Token) []const u8 {
		return switch (self) {
			.end => "end",
			.unknown => |unknown| @as([*]const u8, @ptrCast(&unknown))[0 .. 1],
			.newline => "newline",

			.identifier => |identifier| identifier,
			.builtin => |identifier| identifier,

			.symbol_plus => "+",
			.symbol_minus => "-",
			.symbol_asterisk => "*",
			.symbol_forward_slash => "/",
			.symbol_paren_left => "(",
			.symbol_paren_right => ")",
			.symbol_bang => "!",
			.symbol_comma => ",",
			.symbol_at => "@",
			.symbol_brace_left => "{",
			.symbol_brace_right => "}",
			.symbol_bracket_left => "[",
			.symbol_bracket_right => "]",
			.symbol_period => ".",
			.symbol_colon => ":",
			.symbol_less_than => "<",
			.symbol_less_equals => "<=",
			.symbol_greater_than => ">",
			.symbol_greater_equals => ">=",
			.symbol_equals => "=",
			.symbol_double_equals => "==",

			.number => |literal| literal,
			.string => |literal| literal,
			.template_string => |literal| literal,

			.keyword_const => "const",
			.keyword_nil => "nil",
			.keyword_false => "false",
			.keyword_true => "true",
			.keyword_return => "return",
			.keyword_self => "self",
			.keyword_if => "if",
			.keyword_do => "do",
			.keyword_end => "end",
			.keyword_while => "while",
			.keyword_elif => "elif",
			.keyword_else => "else",
			.keyword_var => "var",
			.keyword_vec2 => "vec2",
			.keyword_vec3 => "vec3",
			.keyword_let => "let",
			.keyword_lambda => "lambda",
		};
	}
};

pub const Stream = struct {
	source: []const u8,
	line: Line = .{.number = 1},
	token: Token = .newline,

	pub fn skip_newlines(self: *Stream) void {
		self.step();

		while (self.token == .newline) {
			self.step();
		}
	}

	pub fn step(self: *Stream) void {
		var cursor = @as(usize, 0);

		defer self.source = self.source[cursor ..];

		while (cursor < self.source.len) {
			switch (self.source[cursor]) {
				'#' => {
					cursor += 1;

					while (cursor < self.source.len and self.source[cursor] != '\n') {
						cursor += 1;
					}
				},

				' ', '\t' => cursor += 1,

				'\n' => {
					cursor += 1;
					self.token = .newline;
					self.line.number += 1;

					return;
				},

				'0' ... '9' => {
					const begin = cursor;

					cursor += 1;

					while (cursor < self.source.len) switch (self.source[cursor]) {
						'0' ... '9' => cursor += 1,

						'.' => {
							cursor += 1;

							while (cursor < self.source.len) switch (self.source[cursor]) {
								'0' ... '9' => cursor += 1,
								else => break,
							};

							self.token = .{.number = self.source[begin .. cursor]};

							return;
						},

						else => break,
					};

					self.token = .{.number = self.source[begin .. cursor]};

					return;
				},

				'A' ... 'Z', 'a' ... 'z', '_' => {
					const begin = cursor;

					cursor += 1;

					while (cursor < self.source.len) switch (self.source[cursor]) {
						'0'...'9', 'A'...'Z', 'a'...'z', '_' => cursor += 1,
						else => break,
					};

					const identifier = self.source[begin .. cursor];

					std.debug.assert(identifier.len != 0);

					switch (identifier[0]) {
						'c' => {
							if (coral.are_equal(identifier[1 ..], "onst")) {
								self.token = .keyword_const;

								return;
							}
						},

						'd' => {
							if (coral.are_equal(identifier[1 ..], "o")) {
								self.token = .keyword_do;

								return;
							}
						},

						'e' => {
							if (coral.are_equal(identifier[1 ..], "lse")) {
								self.token = .keyword_else;

								return;
							}

							if (coral.are_equal(identifier[1 ..], "lif")) {
								self.token = .keyword_elif;

								return;
							}

							if (coral.are_equal(identifier[1 ..], "nd")) {
								self.token = .keyword_end;

								return;
							}
						},

						'f' => {
							if (coral.are_equal(identifier[1 ..], "alse")) {
								self.token = .keyword_false;

								return;
							}
						},

						'i' => {
							if (coral.are_equal(identifier[1 ..], "f")) {
								self.token = .keyword_if;

								return;
							}
						},

						'l' => {
							if (coral.are_equal(identifier[1 ..], "ambda")) {
								self.token = .keyword_lambda;

								return;
							}

							if (coral.are_equal(identifier[1 ..], "et")) {
								self.token = .keyword_let;

								return;
							}
						},

						'n' => {
							if (coral.are_equal(identifier[1 ..], "il")) {
								self.token = .keyword_nil;

								return;
							}
						},

						'r' => {
							if (coral.are_equal(identifier[1 ..], "eturn")) {
								self.token = .keyword_return;

								return;
							}
						},

						's' => {
							if (coral.are_equal(identifier[1 ..], "elf")) {
								self.token = .keyword_self;

								return;
							}
						},

						't' => {
							if (coral.are_equal(identifier[1 ..], "rue")) {
								self.token = .keyword_true;

								return;
							}
						},

						'v' => {
							const rest = identifier[1 ..];

							if (coral.are_equal(rest, "ar")) {
								self.token = .keyword_var;

								return;
							}

							if (coral.are_equal(rest, "vec2")) {
								self.token = .keyword_vec2;

								return;
							}

							if (coral.are_equal(rest, "vec3")) {
								self.token = .keyword_vec3;

								return;
							}
						},

						'w' => {
							if (coral.are_equal(identifier[1 ..], "hile")) {
								self.token = .keyword_while;

								return;
							}
						},

						else => {},
					}

					self.token = .{.identifier = identifier};

					return;
				},

				'`' => {
					cursor += 1;

					const begin = cursor;

					while (cursor < self.source.len) switch (self.source[cursor]) {
						'`' => break,
						else => cursor += 1,
					};

					self.token = .{.template_string = self.source[begin .. cursor]};
					cursor += 1;

					return;
				},

				'"' => {
					cursor += 1;

					const begin = cursor;

					while (cursor < self.source.len) switch (self.source[cursor]) {
						'"' => break,
						else => cursor += 1,
					};

					self.token = .{.string = self.source[begin .. cursor]};
					cursor += 1;

					return;
				},

				'{' => {
					self.token = .symbol_brace_left;
					cursor += 1;

					return;
				},

				'}' => {
					self.token = .symbol_brace_right;
					cursor += 1;

					return;
				},

				'[' => {
					self.token = .symbol_bracket_left;
					cursor += 1;

					return;
				},

				']' => {
					self.token = .symbol_bracket_right;
					cursor += 1;

					return;
				},

				',' => {
					self.token = .symbol_comma;
					cursor += 1;

					return;
				},

				'!' => {
					self.token = .symbol_bang;
					cursor += 1;

					return;
				},

				')' => {
					self.token = .symbol_paren_right;
					cursor += 1;

					return;
				},

				'(' => {
					self.token = .symbol_paren_left;
					cursor += 1;

					return;
				},

				'/' => {
					self.token = .symbol_forward_slash;
					cursor += 1;

					return;
				},

				'*' => {
					self.token = .symbol_asterisk;
					cursor += 1;

					return;
				},

				'-' => {
					self.token = .symbol_minus;
					cursor += 1;

					return;
				},

				'+' => {
					self.token = .symbol_plus;
					cursor += 1;

					return;
				},

				':' => {
					self.token = .symbol_colon;
					cursor += 1;

					return;
				},

				'=' => {
					cursor += 1;

					if (cursor < self.source.len) {
						switch (self.source[cursor]) {
							'=' => {
								cursor += 1;
								self.token = .symbol_double_equals;

								return;
							},

							else => {},
						}
					}

					self.token = .symbol_equals;

					return;
				},

				'<' => {
					cursor += 1;

					if (cursor < self.source.len and (self.source[cursor] == '=')) {
						cursor += 1;
						self.token = .symbol_less_equals;

						return;
					}

					self.token = .symbol_less_than;

					return;
				},

				'>' => {
					cursor += 1;

					if (cursor < self.source.len and (self.source[cursor] == '=')) {
						cursor += 1;
						self.token = .symbol_greater_equals;

						return;
					}

					self.token = .symbol_greater_than;

					return;
				},

				'.' => {
					self.token = .symbol_period;
					cursor += 1;

					return;
				},

				'@' => {
					self.token = .symbol_at;
					cursor += 1;

					return;
				},

				else => {
					self.token = .{.unknown = self.source[cursor]};
					cursor += 1;

					return;
				},
			}
		}

		self.token = .end;

		return;
	}
};
