pub const Expr = @import("./tree/Expr.zig");

pub const Stmt = @import("./tree/Stmt.zig");

const coral = @import("coral");

const std = @import("std");

const script = @import("../script.zig");

const tokens = @import("./tokens.zig");

pub const Declaration = struct {
	identifier: []const coral.Byte,

	is: packed struct {
		readonly: bool = false,
		captured: bool = false,
	} = .{},
};

pub const Environment = struct {
	captures: [capture_max]Capture = [_]Capture{.{.declaration_index = 0}} ** capture_max,
	capture_count: u8 = 0,
	declarations: [declaration_max]Declaration = [_]Declaration{.{.identifier = ""}} ** declaration_max,
	declaration_count: u8 = 0,
	argument_count: u8 = 0,
	statement: ?*const Stmt = null,
	enclosing: ?*Environment = null,

	pub const Capture = union (enum) {
		declaration_index: u8,
		capture_index: u8,
	};

	pub const DeclareError = std.mem.Allocator.Error || error {
		DeclarationExists,
	};

	const capture_max = std.math.maxInt(u8);

	const declaration_max = std.math.maxInt(u8);

	pub fn create_enclosed(self: *Environment, root: *Root) std.mem.Allocator.Error!*Environment {
		const environment = try root.arena.allocator().create(Environment);

		environment.* = .{.enclosing = self};

		return environment;
	}

	fn declare(self: *Environment, declaration: Declaration) DeclareError!*const Declaration {
		if (self.declaration_count == self.declarations.len) {
			return error.OutOfMemory;
		}

		{
			var environment = self;

			while (true) {
				var remaining_count = environment.declaration_count;

				while (remaining_count != 0) {
					remaining_count -= 1;

					if (coral.are_equal(environment.declarations[remaining_count].identifier, declaration.identifier)) {
						return error.DeclarationExists;
					}
				}

				environment = environment.enclosing orelse break;
			}
		}

		const declaration_slot = &self.declarations[self.declaration_count];

		declaration_slot.* = declaration;
		self.declaration_count += 1;

		return declaration_slot;
	}

	pub fn declare_argument(self: *Environment, identifier: []const u8) DeclareError!*const Declaration {
		std.debug.assert(self.declaration_count <= self.argument_count);

		defer self.argument_count += 1;

		return self.declare(.{
			.identifier = identifier,
			.is = .{.readonly = true},
		});
	}

	pub fn declare_constant(self: *Environment, identifier: []const u8) DeclareError!*const Declaration {
		return self.declare(.{
			.identifier = identifier,
			.is = .{.readonly = true},
		});
	}

	pub fn declare_variable(self: *Environment, identifier: []const u8) DeclareError!*const Declaration {
		return self.declare(.{.identifier = identifier});
	}

	pub fn resolve_declaration(self: *Environment, identifier: []const u8) std.mem.Allocator.Error!?*const Declaration {
		var environment = self;
		var ancestry = @as(u32, 0);

		while (true) : (ancestry += 1) {
			var remaining_count = environment.declaration_count;

			while (remaining_count != 0) {
				remaining_count -= 1;

				const declaration = &environment.declarations[remaining_count];

				if (coral.are_equal(declaration.identifier, identifier)) {
					if (ancestry != 0) {
						declaration.is.captured = true;
						environment = self;
						ancestry -= 1;

						while (ancestry != 0) : (ancestry -= 1) {
							if (environment.capture_count == environment.captures.len) {
								return error.OutOfMemory;
							}

							environment.captures[environment.capture_count] = .{
								.capture_index = environment.enclosing.?.capture_count
							};

							environment.capture_count += 1;
							environment = environment.enclosing.?;
						}

						environment.captures[environment.capture_count] = .{.declaration_index = remaining_count};
						environment.capture_count += 1;
					}

					return declaration;
				}
			}

			environment = environment.enclosing orelse return null;
		}
	}

	pub fn get_captures(self: *const Environment) []const Capture {
		return self.captures[0 .. self.capture_count];
	}

	pub fn get_declarations(self: *const Environment) []const Declaration {
		return self.declarations[0 .. self.declaration_count];
	}
};

pub fn NodeChain(comptime Value: type) type {
	return struct {
		head: ?*Value = null,
		tail: ?*Value = null,

		pub const Nodes = struct {
			current: ?*const Value,

			pub fn next(self: *Nodes) ?*const Value {
				const current = self.current orelse return null;

				defer self.current = current.next;

				return current;
			}
		};

		const Self = @This();

		pub fn append(self: *Self, value: *Value) void {
			if (self.tail) |node| {
				node.next = value;
				self.tail = value;
			} else {
				self.tail = value;
				self.head = value;
			}
		}

		pub fn nodes(self: *const Self) Nodes {
			return .{.current = self.head};
		}
	};
}

pub const ParseError = std.mem.Allocator.Error || error {
	BadSyntax,
};

pub const Root = struct {
	arena: std.heap.ArenaAllocator,
	environment: Environment,
	error_messages: MessageList,

	const MessageList = coral.Stack([]coral.Byte);

	pub fn report_error(self: *Root, line: tokens.Line, comptime format: []const u8, args: anytype) ParseError {
		const allocator = self.arena.allocator();
		const message = try coral.utf8.alloc_formatted(allocator, format, args);

		defer allocator.free(message);

		try self.error_messages.push(try coral.utf8.alloc_formatted(allocator, "{line_number}: {message}", .{
			.message = message,
			.line_number = line.number,
		}));

		return error.BadSyntax;
	}

	pub fn report_declare_error(self: *Root, line: tokens.Line, identifier: []const u8, @"error": Environment.DeclareError) ParseError {
		return switch (@"error") {
			error.OutOfMemory => error.OutOfMemory,

			error.DeclarationExists => self.report_error(line, "declaration `{identifier}` already exists", .{
				.identifier = identifier,
			}),
		};
	}

	pub fn create_node(self: *Root, node: anytype) std.mem.Allocator.Error!*@TypeOf(node) {
		const copy = try self.arena.allocator().create(@TypeOf(node));

		copy.* = node;

		return copy;
	}

	pub fn create_string(self: *Root, comptime format: []const u8, args: anytype) std.mem.Allocator.Error![]const u8 {
		return coral.utf8.alloc_formatted(self.arena.allocator(), format, args);
	}

	pub fn deinit(self: *Root) void {
		self.error_messages.deinit();
		self.arena.deinit();
	}

	pub fn init(allocator: std.mem.Allocator) std.mem.Allocator.Error!Root {
		return .{
			.arena = std.heap.ArenaAllocator.init(allocator),
			.error_messages = .{.allocator = allocator},
			.environment = .{},
		};
	}

	pub fn parse(self: *Root, stream: *tokens.Stream) ParseError!void {
		stream.skip_newlines();

		const first_statement = try Stmt.parse(self, stream, &self.environment);
		var current_statement = first_statement;

		while (stream.token != .end) {
			const next_statement = try Stmt.parse(self, stream, &self.environment);

			current_statement.next = next_statement;
			current_statement = next_statement;
		}

		self.environment.statement = first_statement;
	}
};

