const Stmt = @import("./Stmt.zig");

const coral = @import("coral");

const tokens = @import("../tokens.zig");

const tree = @import("../tree.zig");

const std = @import("std");

next: ?*const Self = null,
line: tokens.Line,
kind: Kind,

pub const BinaryOp = struct {
	rhs_operand: *Self,
	lhs_operand: *Self,
	operation: Operation,

	pub const Operation = enum {
		addition,
		subtraction,
		multiplication,
		divsion,
		equals_comparison,
		greater_than_comparison,
		greater_equals_comparison,
		less_than_comparison,
		less_equals_comparison,
	};

	fn parser(comptime parse_next: Parser, comptime operations: []const BinaryOp.Operation) Parser {
		const BinaryOpParser = struct {
			fn parse(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
				var expression = try parse_next(root, stream, environment);

				inline for (operations) |operation| {
					const token = comptime @as(tokens.Token, switch (operation) {
						.addition => .symbol_plus,
						.subtraction => .symbol_minus,
						.multiplication => .symbol_asterisk,
						.divsion => .symbol_forward_slash,
						.equals_comparison => .symbol_double_equals,
						.greater_than_comparison => .symbol_greater_than,
						.greater_equals_comparison => .symbol_greater_equals,
						.less_than_comparison => .symbol_less_than,
						.less_equals_comparison => .symbol_less_equals,
					});

					if (stream.token == std.meta.activeTag(token)) {
						stream.step();

						if (stream.token == .end) {
							return root.report_error(stream.line, "expected other half of expression after `" ++ comptime token.text() ++ "`", .{});
						}

						// TODO: Remove once Zig has fixed struct self-reassignment.
						const unnecessary_temp = expression;

						expression = try root.create_node(Self{
							.line = stream.line,

							.kind = .{
								.binary_op = .{
									.rhs_operand = try parse_next(root, stream, environment),
									.operation = operation,
									.lhs_operand = unnecessary_temp,
								},
							},
						});
					}
				}

				return expression;
			}
		};

		return BinaryOpParser.parse;
	}
};

pub const DeclarationGet = struct {
	declaration: *const tree.Declaration,
};

pub const DeclarationSet = struct {
	declaration: *const tree.Declaration,
	assign: *const Self,
};

pub const FieldGet = struct {
	identifier: []const coral.Byte,
	object: *const Self,
};

pub const FieldSet = struct {
	identifier: []const coral.Byte,
	object: *const Self,
	assign: *const Self,
};

pub const Invoke = struct {
	argument: ?*const Self,
	object: *const Self,
};

const Kind = union (enum) {
	nil_literal,
	true_literal,
	false_literal,
	number_literal: []const u8,
	string_literal: []const u8,
	string_template,
	symbol_literal: []const u8,
	vector2: Vector2,
	vector3: Vector3,
	table: tree.NodeChain(TableEntry),
	group: *Self,
	lambda_construct: LambdaConstruct,
	declaration_get: DeclarationGet,
	declaration_set: DeclarationSet,
	field_get: FieldGet,
	field_set: FieldSet,
	external_get: ExternalGet,
	subscript_get: SubscriptGet,
	subscript_set: SubscriptSet,
	binary_op: BinaryOp,
	unary_op: UnaryOp,
	invoke: Invoke,
};

pub const LambdaConstruct = struct {
	environment: *const tree.Environment,
};

const Parser = fn (root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self;

const ExternalGet = struct {
	name: []const u8,
};

const Self = @This();

pub const SubscriptGet = struct {
	index: *const Self,
	object: *const Self,
};

pub const SubscriptSet = struct {
	index: *const Self,
	object: *const Self,
	assign: *const Self,
};

pub const TableEntry = struct {
	next: ?*const TableEntry = null,
	key: *const Self,
	value: *const Self,
};

const TemplateToken = union (enum) {
	invalid: []const coral.Byte,
	literal: []const coral.Byte,
	expression: []const coral.Byte,

	fn extract(source: *[]const coral.Byte) ?TemplateToken {
		var cursor = @as(usize, 0);

		defer source.* = source.*[cursor ..];

		while (cursor < source.len) {
			switch (source.*[cursor]) {
				'{' => {
					cursor += 1;

					while (true) : (cursor += 1) {
						if (cursor == source.len) {
							return .{.invalid = source.*[0 .. cursor]};
						}

						if (source.*[cursor] == '}') {
							const token = TemplateToken{.expression = source.*[1 .. cursor]};

							cursor += 1;

							return token;
						}
					}
				},

				else => {
					cursor += 1;

					while (true) : (cursor += 1) {
						if (cursor == source.len) {
							return .{.literal = source.*[0 .. cursor]};
						}

						if (source.*[cursor] == '{') {
							const cursor_next = cursor + 1;

							if (cursor_next == source.len) {
								return .{.invalid = source.*[0 .. cursor]};
							}

							if (source.*[cursor_next] == '{') {
								cursor = cursor_next;

								return .{.literal = source.*[0 .. cursor]};
							}

							return .{.literal = source.*[0 .. cursor]};
						}
					}
				}
			}
		}

		return null;
	}
};

pub const UnaryOp = struct {
	operand: *Self,
	operation: Operation,

	pub const Operation = enum {
		numeric_negation,
		boolean_negation,
	};
};

pub const Vector2 = struct {
	x: *const Self,
	y: *const Self,
};

pub const Vector3 = struct {
	x: *const Self,
	y: *const Self,
	z: *const Self,
};

pub fn parse(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
	const expression = try parse_additive(root, stream, environment);

	if (stream.token == .symbol_equals) {
		stream.skip_newlines();

		if (stream.token == .end) {
			return root.report_error(stream.line, "expected assignment after `=`", .{});
		}

		return root.create_node(Self{
			.line = stream.line,

			.kind = switch (expression.kind) {
				.declaration_get => |declaration_get| convert: {
					if (declaration_get.declaration.is.readonly) {
						return root.report_error(stream.line, "readonly declarations cannot be re-assigned", .{});
					}

					break: convert .{
						.declaration_set = .{
							.assign = try parse(root, stream, environment),
							.declaration = declaration_get.declaration,
						},
					};
				},

				.field_get => |field_get| .{
					.field_set = .{
						.assign = try parse(root, stream, environment),
						.object = field_get.object,
						.identifier = field_get.identifier,
					},
				},

				.subscript_get => |subscript_get| .{
					.subscript_set = .{
						.assign = try parse(root, stream, environment),
						.object = subscript_get.object,
						.index = subscript_get.index,
					},
				},

				else => return root.report_error(stream.line, "expected local or field on left-hand side of expression", .{}),
			},
		});
	}

	return expression;
}

const parse_additive = BinaryOp.parser(parse_equality, &.{
	.addition,
	.subtraction,
});

const parse_comparison = BinaryOp.parser(parse_term, &.{
	.greater_than_comparison,
	.greater_equals_comparison,
	.less_than_comparison,
	.less_equals_comparison
});

const parse_equality = BinaryOp.parser(parse_comparison, &.{
	.equals_comparison,
});

fn parse_factor(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
	var expression = try parse_operand(root, stream, environment);

	while (true) {
		switch (stream.token) {
			.symbol_period => {
				stream.skip_newlines();

				// TODO: Remove when Zig fixes miscompilation with in-place struct re-assignment.
				const unnecessary_temp = expression;

				expression = try root.create_node(Self{
					.line = stream.line,

					.kind = .{
						.field_get = .{
							.identifier = switch (stream.token) {
								.identifier => |field_identifier| field_identifier,
								else => return root.report_error(stream.line, "expected identifier after `.`", .{}),
							},

							.object = unnecessary_temp,
						},
					},
				});

				stream.skip_newlines();
			},

			.symbol_bracket_left => {
				stream.skip_newlines();

				// TODO: Remove when Zig fixes miscompilation with in-place struct re-assignment.
				const unnecessary_temp = expression;

				expression = try root.create_node(Self{
					.line = stream.line,

					.kind = .{
						.subscript_get = .{
							.index = try parse(root, stream, environment),
							.object = unnecessary_temp,
						},
					},
				});

				if (stream.token != .symbol_bracket_right) {
					return root.report_error(stream.line, "expected closing `]` on subscript", .{});
				}

				stream.skip_newlines();
			},

			.symbol_paren_left => {
				const lines_stepped = stream.line;

				stream.skip_newlines();

				var first_argument = @as(?*Self, null);

				if (stream.token != .symbol_paren_right) {
					var argument = try parse(root, stream, environment);

					first_argument = argument;

					while (true) {
						switch (stream.token) {
							.symbol_comma => stream.skip_newlines(),
							.symbol_paren_right => break,
							else => return root.report_error(stream.line, "expected `,` or `)` after lambda argument", .{}),
						}

						const next_argument = try parse(root, stream, environment);

						argument.next = next_argument;
						argument = next_argument;
					}
				}

				stream.skip_newlines();

				// TODO: Remove when Zig fixes miscompilation with in-place struct re-assignment.
				const unnecessary_temp = expression;

				expression = try root.create_node(Self{
					.line = lines_stepped,

					.kind = .{
						.invoke = .{
							.argument = first_argument,
							.object = unnecessary_temp,
						},
					},
				});
			},

			else => break,
		}
	}

	return expression;
}

fn parse_operand(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
	switch (stream.token) {
		.symbol_paren_left => {
			stream.skip_newlines();

			const expression = try parse(root, stream, environment);

			if (stream.token != .symbol_paren_right) {
				return root.report_error(stream.line, "expected a closing `)` after expression", .{});
			}

			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .{.group = expression},
			});
		},

		.keyword_nil => {
			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .nil_literal,
			});
		},

		.keyword_true => {
			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .true_literal,
			});
		},

		.keyword_false => {
			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .false_literal,
			});
		},

		.keyword_vec2 => {
			stream.skip_newlines();

			if (stream.token != .symbol_paren_left) {
				return root.report_error(stream.line, "expected an opening `(` after `vec2`", .{});
			}

			stream.skip_newlines();

			const x_expression = try parse(root, stream, environment);

			switch (stream.token) {
				.symbol_paren_right => {
					stream.skip_newlines();

					return root.create_node(Self{
						.line = stream.line,

						.kind = .{
							.vector2 = .{
								.x = x_expression,
								.y = x_expression,
							},
						},
					});
				},

				.symbol_comma => {
					stream.skip_newlines();

					const y_expression = try parse(root, stream, environment);

					stream.skip_newlines();

					if (stream.token != .symbol_paren_right) {
						return root.report_error(stream.line, "expected a closing `)` after `vec3`", .{});
					}

					return root.create_node(Self{
						.line = stream.line,

						.kind = .{
							.vector2 = .{
								.x = x_expression,
								.y = y_expression,
							},
						},
					});
				},

				else => return root.report_error(stream.line, "expected a closing `)` after `vec3`", .{}),
			}
		},

		.keyword_vec3 => {
			stream.skip_newlines();

			if (stream.token != .symbol_paren_left) {
				return root.report_error(stream.line, "expected an opening `(` after `vec2`", .{});
			}

			stream.skip_newlines();

			const x_expression = try parse(root, stream, environment);

			switch (stream.token) {
				.symbol_paren_right => {
					stream.skip_newlines();

					return root.create_node(Self{
						.line = stream.line,

						.kind = .{
							.vector3 = .{
								.x = x_expression,
								.y = x_expression,
								.z = x_expression,
							},
						},
					});
				},

				.symbol_comma => {
					stream.skip_newlines();

					const y_expression = try parse(root, stream, environment);

					stream.skip_newlines();

					const z_expression = try parse(root, stream, environment);

					stream.skip_newlines();

					if (stream.token != .symbol_paren_right) {
						return root.report_error(stream.line, "expected a closing `)` after `vec3`", .{});
					}

					return root.create_node(Self{
						.line = stream.line,

						.kind = .{
							.vector3 = .{
								.x = x_expression,
								.y = y_expression,
								.z = z_expression,
							},
						},
					});
				},

				else => return root.report_error(stream.line, "expected a closing `)` after `vec3`", .{}),
			}
		},

		.number => |value| {
			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .{.number_literal = value},
			});
		},

		.string => |value| {
			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .{.string_literal = value},
			});
		},

		.template_string => |value| {
			const line = stream.line;

			stream.skip_newlines();

			return parse_template(root, value, line, environment);
		},

		.symbol_at => {
			stream.step();

			const identifier = switch (stream.token) {
				.identifier => |identifier| identifier,
				else => return root.report_error(stream.line, "expected identifier after `@`", .{}),
			};

			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .{.external_get = .{.name = identifier}},
			});
		},

		.symbol_period => {
			stream.step();

			const identifier = switch (stream.token) {
				.identifier => |identifier| identifier,
				else => return root.report_error(stream.line, "expected identifier after `.`", .{}),
			};

			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .{.symbol_literal = identifier},
			});
		},

		.identifier => |identifier| {
			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,

				.kind = .{
					.declaration_get = .{
						.declaration = (try environment.resolve_declaration(identifier)) orelse {
							return root.report_error(stream.line, "undefined identifier `{identifier}`", .{
								.identifier = identifier,
							});
						}
					},
				},
			});
		},

		.keyword_lambda => {
			stream.skip_newlines();

			if (stream.token != .symbol_paren_left) {
				return root.report_error(stream.line, "expected `(` after opening lambda block", .{});
			}

			stream.skip_newlines();

			var lambda_environment = try environment.create_enclosed(root);

			while (stream.token != .symbol_paren_right) {
				const identifier = switch (stream.token) {
					.identifier => |identifier| identifier,
					else => return root.report_error(stream.line, "expected identifier", .{}),
				};

				_ = lambda_environment.declare_argument(identifier) catch |declare_error| {
					return root.report_declare_error(stream.line, identifier, declare_error);
				};

				stream.skip_newlines();

				switch (stream.token) {
					.symbol_comma => stream.skip_newlines(),
					.symbol_paren_right => break,
					else => return root.report_error(stream.line, "expected `,` or `)` after identifier", .{}),
				}
			}

			stream.skip_newlines();

			if (stream.token != .symbol_colon) {
				return root.report_error(stream.line, "expected `:` after closing `)` of lambda identifiers", .{});
			}

			stream.skip_newlines();

			if (stream.token != .keyword_end) {
				const first_statement = try Stmt.parse(root, stream, lambda_environment);
				var current_statement = first_statement;

				while (stream.token != .keyword_end) {
					const next_statement = try Stmt.parse(root, stream, lambda_environment);

					current_statement.next = next_statement;
					current_statement = next_statement;
				}

				lambda_environment.statement = first_statement;
			}

			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,
				.kind = .{.lambda_construct = .{.environment = lambda_environment}},
			});
		},

		.symbol_brace_left => {
			stream.skip_newlines();

			return parse_table(root, stream, environment);
		},

		.symbol_minus => {
			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,

				.kind = .{
					.unary_op = .{
						.operand = try parse_factor(root, stream, environment),
						.operation = .numeric_negation,
					},
				},
			});
		},

		.symbol_bang => {
			stream.skip_newlines();

			return root.create_node(Self{
				.line = stream.line,

				.kind = .{
					.unary_op = .{
						.operand = try parse_factor(root, stream, environment),
						.operation = .boolean_negation,
					},
				},
			});
		},

		else => return root.report_error(stream.line, "unexpected token in expression", .{}),
	}
}

fn parse_table(root: *tree.Root, stream: *tokens.Stream, environment: *tree.Environment) tree.ParseError!*Self {
	var entries = tree.NodeChain(TableEntry){};
	var sequential_index = @as(usize, 0);

	while (true) {
		switch (stream.token) {
			.symbol_brace_right => {
				stream.skip_newlines();

				return root.create_node(Self{
					.line = stream.line,
					.kind = .{.table = entries},
				});
			},

			.symbol_bracket_left => {
				stream.skip_newlines();

				const key = try parse(root, stream, environment);

				if (stream.token != .symbol_bracket_right)  {
					return root.report_error(stream.line, "expected `]` after subscript index expression", .{});
				}

				stream.skip_newlines();

				if (stream.token != .symbol_equals) {
					return root.report_error(stream.line, "expected `=` after table expression key", .{});
				}

				stream.skip_newlines();

				entries.append(try root.create_node(TableEntry{
					.value = try parse(root, stream, environment),
					.key = key,
				}));
			},

			.symbol_period => {
				stream.step();

				const field = switch (stream.token) {
					.identifier => |identifier| identifier,
					else => return root.report_error(stream.line, "invalid symbol literal", .{}),
				};

				stream.skip_newlines();

				switch (stream.token) {
					.symbol_comma => {
						stream.skip_newlines();

						entries.append(try root.create_node(TableEntry{
							.key = try root.create_node(Self{
								.line = stream.line,
								.kind = .{.number_literal = try root.create_string("{i}", .{.i = sequential_index})},
							}),

							.value = try root.create_node(Self{
								.line = stream.line,
								.kind = .{.symbol_literal = field},
							}),
						}));

						sequential_index += 1;
					},

					.symbol_equals => {
						stream.skip_newlines();

						entries.append(try root.create_node(TableEntry{
							.value = try parse(root, stream, environment),

							.key = try root.create_node(Self{
								.line = stream.line,
								.kind = .{.symbol_literal = field},
							}),
						}));
					},

					else => return root.report_error(stream.line, "expected `,` or `=` after symbol", .{}),
				}
			},

			else => {
				entries.append(try root.create_node(TableEntry{
					.value = try parse(root, stream, environment),

					.key = try root.create_node(Self{
						.line = stream.line,
						.kind = .{.number_literal = try root.create_string("{i}", .{.i = sequential_index})},
					}),
				}));

				sequential_index += 1;
			},
		}

		switch (stream.token) {
			.symbol_brace_right => {
				stream.skip_newlines();

				return root.create_node(Self{
					.line = stream.line,
					.kind = .{.table = entries},
				});
			},

			.symbol_comma => stream.skip_newlines(),
			else => return root.report_error(stream.line, "expected `,` or '}' after table key value pair", .{}),
		}
	}
}

fn parse_template(root: *tree.Root, template: []const coral.Byte, line: tokens.Line, environment: *tree.Environment) tree.ParseError!*Self {
	const expression_head = try root.create_node(Self{
		.line = line,
		.kind = .string_template,
	});

	var expression_tail = expression_head;
	var source = template;

	while (TemplateToken.extract(&source)) |token| {
		const expression = try switch (token) {
			.invalid => |invalid| root.report_error(line, "invalid template format: `{invalid}`", .{
				.invalid = invalid,
			}),

			.literal => |literal| root.create_node(Self{
				.line = line,
				.kind = .{.string_literal = literal},
			}),

			.expression => |expression| create: {
				var stream = tokens.Stream{
					.source = expression,
					.line = line,
				};

				stream.step();

				break: create try parse(root, &stream, environment);
			},
		};

		expression_tail.next = expression;
		expression_tail = expression;
	}

	return expression_head;
}

const parse_term = BinaryOp.parser(parse_factor, &.{
	.multiplication,
	.divsion,
});
