const App = @import("./App.zig");

const coral = @import("coral");

const device = @import("./gfx/device.zig");

const ext = @import("./ext.zig");

const formats = @import("./gfx/formats.zig");

const msg = @import("./msg.zig");

const std = @import("std");

pub const Assets = struct {
	context: device.Context,
	formats: coral.stack.Sequential(Format),
	staging_arena: std.heap.ArenaAllocator,

	pub const Format = struct {
		extension: []const u8,
		file_desc: *const fn (*std.heap.ArenaAllocator, []const u8) Error!Desc,

		pub const Error = std.mem.Allocator.Error || coral.files.Error || error {
			FormatUnsupported,
		};
	};

	pub fn open_file(self: *Assets, path: []const u8) (std.mem.Allocator.Error || Format.Error)!Handle {
		defer {
			const max_cache_size = 536870912;

			if (!self.staging_arena.reset(.{.retain_with_limit = max_cache_size})) {
				std.log.warn("failed to retain staging arena size of {} bytes", .{max_cache_size});
			}
		}

		for (self.formats.values) |format| {
			if (!std.mem.endsWith(u8, path, format.extension)) {
				continue;
			}

			return self.context.open(try format.file_desc(&self.staging_arena, path));
		}

		return .none;
	}

	pub fn open_quad_mesh_2d(self: *Assets, extents: Point2D) std.mem.Allocator.Error!Handle {
		const width, const height = extents / @as(Point2D, @splat(2));

		return self.context.open(.{
			.mesh_2d = .{
				.indices = &.{0, 1, 2, 0, 2, 3},

				.vertices = &.{
					.{.xy = .{-width, height}, .uv = .{0, 1}},
					.{.xy = .{width, height}, .uv = .{1, 1}},
					.{.xy = .{width, -height}, .uv = .{1, 0}},
					.{.xy = .{-width, -height},  .uv = .{0, 0}},
				},
			},
		});
	}
};

pub const Color = @Vector(4, f32);

pub const Desc = union (enum) {
	texture: Texture,
	mesh_2d: Mesh2D,

	pub const Mesh2D = struct {
		vertices: []const Vertex,
		indices: []const u16,

		pub const Vertex = struct {
			xy: Point2D,
			uv: Point2D,
		};
	};

	pub const Texture = struct {
		data: []const coral.io.Byte,
		width: u16,
		format: Format,
		access: Access,

		pub const Access = enum {
			static,
		};

		pub const Format = enum {
			rgba8,
			bgra8,

			pub fn byte_size(self: Format) usize {
				return switch (self) {
					.rgba8, .bgra8 => 4,
				};
			}
		};
	};
};

pub const Display = struct {
	width: u16 = 1280,
	height: u16 = 720,
	clear_color: Color = colors.black,
};

pub const Handle = enum (usize) {
	none,
	_,

	pub fn index(self: Handle) ?usize {
		return switch (self) {
			.none => null,
			_ => @intFromEnum(self) - 1,
		};
	}
};

pub const Input = union (enum) {
	key_up: Key,
	key_down: Key,

	pub const Key = enum (u32) {
		a = ext.SDL_SCANCODE_A,
		d = ext.SDL_SCANCODE_D,
		s = ext.SDL_SCANCODE_S,
		w = ext.SDL_SCANCODE_W,

		pub fn scancode(key: Key) u32 {
			return @intFromEnum(key);
		}
	};
};

pub const Point2D = @Vector(2, f32);

pub const Queue = struct {
	commands: *device.RenderList,

	pub const State = struct {
		command_index: usize,
	};

	pub fn bind(_: coral.system.BindContext) std.mem.Allocator.Error!State {
		// TODO: Review how good of an idea this global state is, even if bind is guaranteed to always be ran on main.
		if (renders.is_empty()) {
			renders = .{.allocator = coral.heap.allocator};
		}

		const command_index = renders.len();

		try renders.push_grow(device.RenderChain.init(coral.heap.allocator));

		return .{
			.command_index = command_index,
		};
	}

	pub fn init(state: *State) Queue {
		return .{
			.commands = renders.values[state.command_index].pending(),
		};
	}

	pub fn unbind(state: *State) void {
		std.debug.assert(!renders.is_empty());

		const render = &renders.values[state.command_index];

		render.deinit();
		std.mem.swap(device.RenderChain, render, renders.get_ptr().?);
		std.debug.assert(renders.pop());

		if (renders.is_empty()) {
			Queue.renders.deinit();
		}
	}

	var renders = coral.stack.Sequential(device.RenderChain){.allocator = coral.heap.allocator};
};

pub const Transform2D = extern struct {
	xbasis: Point2D = .{1, 0},
	ybasis: Point2D = .{0, 1},
	origin: Point2D = @splat(0),
};

const builtin_formats = [_]Assets.Format{
	.{
		.extension = "bmp",
		.file_desc = formats.bmp_file_desc,
	},
};

pub const colors = struct {
	pub const black = greyscale(0);

	pub const white = greyscale(1);

	pub fn greyscale(v: f32) Color {
		return .{v, v, v, 1};
	}

	pub fn rgb(r: f32, g: f32, b: f32) Color {
		return .{r, g, b, 1};
	}
};

pub fn poll(app: coral.Write(App), inputs: msg.Send(Input)) !void {
	var event = @as(ext.SDL_Event, undefined);

	while (ext.SDL_PollEvent(&event) != 0) {
		switch (event.type) {
			ext.SDL_QUIT => app.res.quit(),
			ext.SDL_KEYUP => try inputs.push(.{.key_up = @enumFromInt(event.key.keysym.scancode)}),
			ext.SDL_KEYDOWN => try inputs.push(.{.key_down = @enumFromInt(event.key.keysym.scancode)}),
			else => {},
		}
	}
}

pub fn setup(world: *coral.World, events: App.Events) (error {Unsupported} || std.Thread.SpawnError || std.mem.Allocator.Error)!void {
	if (ext.SDL_Init(ext.SDL_INIT_VIDEO) != 0) {
		return error.Unsupported;
	}

	var context = try device.Context.init();

	errdefer context.deinit();

	var registered_formats = coral.stack.Sequential(Assets.Format){.allocator = coral.heap.allocator};

	errdefer registered_formats.deinit();

	try registered_formats.grow(builtin_formats.len);
	std.debug.assert(registered_formats.push_all(&builtin_formats));

	try world.set_resource(.none, Assets{
		.staging_arena = std.heap.ArenaAllocator.init(coral.heap.allocator),
		.formats = registered_formats,
		.context = context,
	});

	try world.set_resource(.none, Display{});
	try world.on_event(events.pre_update, coral.system_fn(poll), .{.label = "poll gfx"});
	try world.on_event(events.exit, coral.system_fn(stop), .{.label = "stop gfx"});
	try world.on_event(events.finish, coral.system_fn(synchronize), .{.label = "synchronize gfx"});
}

pub fn stop(assets: coral.Write(Assets)) void {
	assets.res.staging_arena.deinit();
	assets.res.formats.deinit();
	assets.res.context.deinit();
}

pub fn synchronize(assets: coral.Write(Assets), display: coral.Read(Display)) !void {
	assets.res.context.submit(.{
		.width = display.res.width,
		.height = display.res.height,
		.clear_color = display.res.clear_color,
		.renders = Queue.renders.values,
	});
}
