const commands = @import("./commands.zig");

const coral = @import("coral");

const ext = @import("../ext.zig");

const gfx = @import("../gfx.zig");

const sokol = @import("sokol");

const std = @import("std");

pub const Context = struct {
	window: *ext.SDL_Window,
	thread: std.Thread,
	loop: *Loop,

	pub const Submission = struct {
		width: u16,
		height: u16,
		clear_color: gfx.Color = gfx.colors.black,
		renders: []RenderChain,
	};

	pub fn close(self: *Context, handle: gfx.Handle) void {
		const handle_index = handle.index() orelse {
			return;
		};

		const close_commands = self.loop.closes.pending();

		std.debug.assert(close_commands.stack.cap > close_commands.stack.len());
		close_commands.append(.{.index = handle_index}) catch unreachable;
	}

	pub fn deinit(self: *Context) void {
		self.loop.is_running.store(false, .monotonic);
		self.loop.ready.post();
		self.thread.join();
		self.loop.deinit();
		coral.heap.allocator.destroy(self.loop);
		ext.SDL_DestroyWindow(self.window);

		self.* = undefined;
	}

	pub fn init() !Context {
		const window = create: {
			const position = ext.SDL_WINDOWPOS_CENTERED;
			const flags = ext.SDL_WINDOW_OPENGL;
			const width = 640;
			const height = 480;

			break: create ext.SDL_CreateWindow("Ona", position, position, width, height, flags) orelse {
				return error.Unsupported;
			};
		};

		errdefer ext.SDL_DestroyWindow(window);

		const loop = try coral.heap.allocator.create(Loop);

		errdefer coral.heap.allocator.destroy(loop);

		loop.* = .{};

		return .{
			.loop = loop,
			.window = window,

			.thread = spawn: {
				const thread = try std.Thread.spawn(.{}, Loop.run, .{loop, window});

				thread.setName("Ona Graphics") catch {
					std.log.warn("failed to name the graphics thread", .{});
				};

				break: spawn thread;
			},
		};
	}

	pub fn open(self: *Context, desc: gfx.Desc) std.mem.Allocator.Error!gfx.Handle {
		const open_commands = self.loop.opens.pending();
		const index = self.loop.closed_indices.get() orelse open_commands.stack.len();

		try open_commands.append(.{
			.index = index,
			.desc = desc,
		});

		const pending_closes = self.loop.closes.pending();

		if (pending_closes.stack.len() == pending_closes.stack.cap) {
			try pending_closes.stack.grow(1);
		}

		_ = self.loop.closed_indices.pop();

		return @enumFromInt(index + 1);
	}

	pub fn submit(self: *Context, submission: Submission) void {
		self.loop.finished.wait();

		defer self.loop.ready.post();

		for (submission.renders) |*render| {
			render.swap();
		}

		self.loop.opens.swap();
		self.loop.closes.swap();

		var last_width, var last_height = [_]c_int{0, 0};

		ext.SDL_GetWindowSize(self.window, &last_width, &last_height);

		if (submission.width != last_width or submission.height != last_height) {
			ext.SDL_SetWindowSize(self.window, submission.width, submission.height);
		}

		self.loop.clear_color = submission.clear_color;
		self.loop.renders = submission.renders;

		self.loop.ready.post();
	}
};

const Loop = struct {
	ready: std.Thread.Semaphore = .{},
	finished: std.Thread.Semaphore = .{},
	clear_color: gfx.Color = gfx.colors.black,
	is_running: AtomicBool = AtomicBool.init(true),
	renders: []RenderChain = &.{},
	closes: CloseChain = CloseChain.init(coral.heap.allocator),
	opens: OpenChain = OpenChain.init(coral.heap.allocator),
	closed_indices: coral.stack.Sequential(usize) = .{.allocator = coral.heap.allocator},

	const AtomicBool = std.atomic.Value(bool);

	const CloseCommand = struct {
		index: usize,
	};

	const OpenCommand = struct {
		index: usize,
		desc: gfx.Desc,

		fn clone(command: OpenCommand, arena: *std.heap.ArenaAllocator) std.mem.Allocator.Error!OpenCommand {
			const allocator = arena.allocator();

			return .{
				.desc = switch (command.desc) {
					.texture => |texture| .{
						.texture = .{
							.data = try allocator.dupe(coral.io.Byte, texture.data),
							.width = texture.width,
							.format = texture.format,
							.access = texture.access,
						},
					},

					.mesh_2d => |mesh_2d| .{
						.mesh_2d = .{
							.vertices = try allocator.dupe(gfx.Desc.Mesh2D.Vertex, mesh_2d.vertices),
							.indices = try allocator.dupe(u16, mesh_2d.indices),
						},
					},
				},

				.index = command.index,
			};
		}
	};

	const CloseChain = commands.Chain(CloseCommand, null);

	const OpenChain = commands.Chain(OpenCommand, OpenCommand.clone);

	fn deinit(self: *Loop) void {
		self.closes.deinit();
		self.opens.deinit();
		self.closed_indices.deinit();
	}

	fn run(self: *Loop, window: *ext.SDL_Window) !void {
		const context = configure_and_create: {
			var result = @as(c_int, 0);

			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_FLAGS, ext.SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG);
			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_PROFILE_MASK, ext.SDL_GL_CONTEXT_PROFILE_CORE);
			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MAJOR_VERSION, 3);
			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MINOR_VERSION, 3);
			result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_DOUBLEBUFFER, 1);

			if (result != 0) {
				return error.Unsupported;
			}

			break: configure_and_create ext.SDL_GL_CreateContext(window);
		};

		sokol.gfx.setup(.{
			.environment = .{
				.defaults = .{
					.color_format = .RGBA8,
					.depth_format = .DEPTH_STENCIL,
					.sample_count = 1,
				},
			},

			.logger = .{
				.func = sokol.log.func,
			},
		});

		defer {
			sokol.gfx.shutdown();
			ext.SDL_GL_DeleteContext(context);
		}

		var rendering = Rendering.init();

		defer rendering.deinit();

		self.finished.post();

		while (self.is_running.load(.monotonic)) {
			self.ready.wait();

			defer self.finished.post();

			const open_commands = self.opens.submitted();

			defer open_commands.clear();

			for (open_commands.stack.values) |command| {
				switch (command.desc) {
					.texture => |texture| {
						const stride = texture.width * texture.format.byte_size();

						const image = sokol.gfx.makeImage(.{
							.width = texture.width,
							.height = @intCast(texture.data.len / stride),

							.data = .{
								.subimage = get: {
									var subimage = [_][16]sokol.gfx.Range{.{.{}} ** 16} ** 6;

									subimage[0][0] = sokol.gfx.asRange(texture.data);

									break: get subimage;
								},
							},

							.pixel_format = switch (texture.format) {
								.rgba8 => .RGBA8,
								.bgra8 => .BGRA8,
							},
						});

						errdefer sokol.gfx.destroyImage(image);

						const sampler = sokol.gfx.makeSampler(.{});

						errdefer sokol.gfx.destroySampler(sampler);

						try rendering.insert_object(command.index, .{
							.texture = .{
								.sampler = sampler,
								.image = image,
							},
						});
					},

					.mesh_2d => |mesh_2d| {
						const index_buffer = sokol.gfx.makeBuffer(.{
							.data = sokol.gfx.asRange(mesh_2d.indices),
							.type = .INDEXBUFFER,
						});

						const vertex_buffer = sokol.gfx.makeBuffer(.{
							.data = sokol.gfx.asRange(mesh_2d.vertices),
							.type = .VERTEXBUFFER,
						});

						errdefer {
							sokol.gfx.destroyBuffer(index_buffer);
							sokol.gfx.destroyBuffer(vertex_buffer);
						}

						if (mesh_2d.indices.len > std.math.maxInt(u32)) {
							return error.OutOfMemory;
						}

						try rendering.insert_object(command.index, .{
							.mesh_2d = .{
								.index_buffer = index_buffer,
								.vertex_buffer = vertex_buffer,
								.index_count = @intCast(mesh_2d.indices.len),
							},
						});
					},
				}
			}

			var frame = init_frame: {
				var width, var height = [_]c_int{0, 0};

				ext.SDL_GL_GetDrawableSize(window, &width, &height);
				std.debug.assert(width > 0 and height > 0);

				break: init_frame Rendering.Frame{
					.width = @intCast(width),
					.height = @intCast(height),
				};
			};

			sokol.gfx.beginPass(.{
				.swapchain = .{
					.width = frame.width,
					.height = frame.height,
					.sample_count = 1,
					.color_format = .RGBA8,
					.depth_format = .DEPTH_STENCIL,
					.gl = .{.framebuffer = 0},
				},

				.action = .{
					.colors = get: {
						var actions = [_]sokol.gfx.ColorAttachmentAction{.{}} ** 4;

						actions[0] = .{
							.load_action = .CLEAR,
							.clear_value = @as(sokol.gfx.Color, @bitCast(self.clear_color)),
						};

						break: get actions;
					},
				},
			});

			for (self.renders) |*render| {
				const render_commands = render.submitted();

				defer render_commands.clear();

				for (render_commands.stack.values) |command| {
					switch (command) {
						.instance_2d => |instance_2d| {
							try rendering.push_instance_2d(&frame, instance_2d);
						},

						.post_process => |post_process| {
							rendering.flush_instance_2ds(&frame);
							// sokol.gfx.applyPipeline(self.post_process_pipeline);

							_ = post_process;
						},
					}
				}
			}

			rendering.flush_instance_2ds(&frame);
			sokol.gfx.endPass();
			sokol.gfx.commit();
			ext.SDL_GL_SwapWindow(window);

			const close_commands = self.closes.submitted();

			defer close_commands.clear();

			for (close_commands.stack.values) |command| {
				const object = &rendering.objects.values[command.index];

				switch (object.*) {
					.empty => {}, // TODO: Handle double-closes.

					.mesh_2d => |mesh_2d| {
						sokol.gfx.destroyBuffer(mesh_2d.vertex_buffer);
						sokol.gfx.destroyBuffer(mesh_2d.index_buffer);
					},

					.texture => |texture| {
						sokol.gfx.destroyImage(texture.image);
						sokol.gfx.destroySampler(texture.sampler);
					},
				}

				object.* = .empty;
			}
		}
	}
};

const Rendering = struct {
	objects: coral.stack.Sequential(Object),
	instance_2d_pipeline: sokol.gfx.Pipeline,
	instance_2d_buffers: coral.stack.Sequential(sokol.gfx.Buffer),

	const Instance2D = extern struct {
		transform: gfx.Transform2D,
		tint: @Vector(4, u8) = @splat(std.math.maxInt(u8)),
		depth: f32 = 0,
		texture_offset: gfx.Point2D = @splat(0),
		texture_size: gfx.Point2D = @splat(1),

		const buffer_indices = .{
			.mesh = 0,
			.instance = 1,
		};

		const instances_per_buffer = 512;

		const shader = @import("./shaders/instance_2d.glsl.zig");
	};

	const Frame = struct {
		width: u16,
		height: u16,
		flushed_instance_2d_count: usize = 0,
		pushed_instance_2d_count: usize = 0,
		mesh_2d: gfx.Handle = .none,
		texture: gfx.Handle = .none,

		fn unflushed_instance_2d_count(self: Frame) usize {
			return self.pushed_instance_2d_count - self.flushed_instance_2d_count;
		}
	};

	const Object = union (enum) {
		empty,

		mesh_2d: struct {
			index_count: u32,
			vertex_buffer: sokol.gfx.Buffer,
			index_buffer: sokol.gfx.Buffer,
		},

		texture: struct {
			image: sokol.gfx.Image,
			sampler: sokol.gfx.Sampler,
		},
	};

	fn deinit(self: *Rendering) void {
		for (self.instance_2d_buffers.values) |buffer| {
			sokol.gfx.destroyBuffer(buffer);
		}

		self.instance_2d_buffers.deinit();
		sokol.gfx.destroyPipeline(self.instance_2d_pipeline);
		self.objects.deinit();
	}

	fn init() Rendering {
		sokol.gfx.setup(.{
			.environment = .{
				.defaults = .{
					.color_format = .RGBA8,
					.depth_format = .DEPTH_STENCIL,
					.sample_count = 1,
				},
			},

			.logger = .{
				.func = sokol.log.func,
			},
		});

		return .{
			.instance_2d_pipeline = sokol.gfx.makePipeline(.{
				.label = "2D drawing pipeline",

				.layout = .{
					.attrs = get: {
						var attrs = [_]sokol.gfx.VertexAttrState{.{}} ** 16;

						attrs[Instance2D.shader.ATTR_vs_mesh_xy] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.mesh,
						};

						attrs[Instance2D.shader.ATTR_vs_mesh_uv] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.mesh,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_xbasis] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_ybasis] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_origin] = .{
							.format = .FLOAT2,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_tint] = .{
							.format = .UBYTE4N,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_depth] = .{
							.format = .FLOAT,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						attrs[Instance2D.shader.ATTR_vs_instance_rect] = .{
							.format = .FLOAT4,
							.buffer_index = Instance2D.buffer_indices.instance,
						};

						break: get attrs;
					},

					.buffers = get: {
						var buffers = [_]sokol.gfx.VertexBufferLayoutState{.{}} ** 8;

						buffers[Instance2D.buffer_indices.instance].step_func = .PER_INSTANCE;

						break: get buffers;
					},
				},

				.shader = sokol.gfx.makeShader(Instance2D.shader.draw2dShaderDesc(sokol.gfx.queryBackend())),
				.index_type = .UINT16,
			}),

			.instance_2d_buffers = .{.allocator = coral.heap.allocator},
			.objects = .{.allocator = coral.heap.allocator},
		};
	}

	fn flush_instance_2ds(self: *Rendering, frame: *Frame) void {
		const unflushed_count = frame.unflushed_instance_2d_count();

		if (unflushed_count == 0) {
			return;
		}

		sokol.gfx.applyPipeline(self.instance_2d_pipeline);

		sokol.gfx.applyUniforms(.VS, Instance2D.shader.SLOT_Screen, sokol.gfx.asRange(&Instance2D.shader.Screen{
			.screen_size = .{@floatFromInt(frame.width), @floatFromInt(frame.height)},
		}));

		const mesh_2d = self.objects.values[frame.mesh_2d.index().?].mesh_2d;
		const texture = self.objects.values[frame.texture.index().?].texture;

		var bindings = sokol.gfx.Bindings{
			.vertex_buffers = get: {
				var buffers = [_]sokol.gfx.Buffer{.{}} ** 8;

				buffers[Instance2D.buffer_indices.mesh] = mesh_2d.vertex_buffer;

				break: get buffers;
			},

			.index_buffer = mesh_2d.index_buffer,

			.fs = .{
				.images = get: {
					var images = [_]sokol.gfx.Image{.{}} ** 12;

					images[0] = texture.image;

					break: get images;
				},

				.samplers = get: {
					var samplers = [_]sokol.gfx.Sampler{.{}} ** 8;

					samplers[0] = texture.sampler;

					break: get samplers;
				},
			},
		};

		while (frame.flushed_instance_2d_count < frame.pushed_instance_2d_count) {
			const buffer_index = frame.flushed_instance_2d_count / Instance2D.instances_per_buffer;
			const buffer_offset = frame.flushed_instance_2d_count % Instance2D.instances_per_buffer;
			const instances_to_flush = @min(Instance2D.instances_per_buffer - buffer_offset, unflushed_count);

			bindings.vertex_buffers[Instance2D.buffer_indices.instance] = self.instance_2d_buffers.values[buffer_index];
			bindings.vertex_buffer_offsets[Instance2D.buffer_indices.instance] = @intCast(buffer_offset);

			sokol.gfx.applyBindings(bindings);
			sokol.gfx.draw(0, mesh_2d.index_count, @intCast(instances_to_flush));

			frame.flushed_instance_2d_count += instances_to_flush;
		}
	}

	fn insert_object(self: *Rendering, index: usize, object: Object) !void {
		const resource_count = self.objects.len();

		if (index < resource_count) {
			const empty_object = &self.objects.values[index];

			if (empty_object.* != .empty) {
				return error.InvalidHandle;
			}

			empty_object.* = object;
		} else {
			if (index != resource_count) {
				return error.InvalidIndex;
			}

			try self.objects.push_grow(object);
		}
	}

	fn push_instance_2d(self: *Rendering, frame: *Frame, command: RenderCommand.Instance) std.mem.Allocator.Error!void {
		if (command.mesh_2d != frame.mesh_2d or command.texture != frame.texture) {
			self.flush_instance_2ds(frame);
		}

		frame.mesh_2d = command.mesh_2d;
		frame.texture = command.texture;

		const has_filled_buffer = (frame.pushed_instance_2d_count % Instance2D.instances_per_buffer) == 0;
		const pushed_buffer_count = frame.pushed_instance_2d_count / Instance2D.instances_per_buffer;

		if (has_filled_buffer and pushed_buffer_count == self.instance_2d_buffers.len()) {
			const instance_buffer = sokol.gfx.makeBuffer(.{
				.size = @sizeOf(Instance2D) * Instance2D.instances_per_buffer,
				.usage = .STREAM,
				.label = "2D drawing instance buffer",
			});

			errdefer sokol.gfx.destroyBuffer(instance_buffer);

			try self.instance_2d_buffers.push_grow(instance_buffer);
		}

		_ = sokol.gfx.appendBuffer(self.instance_2d_buffers.get().?, sokol.gfx.asRange(&Instance2D{
			.transform = command.transform,
		}));

		frame.pushed_instance_2d_count += 1;
	}

	fn remove_object(self: *Rendering, index: usize) ?Object {
		const object = self.objects.values[index];

		if (object != .empty) {
			self.objects.values[index] = .empty;

			return object;
		}

		return null;
	}
};

pub const RenderCommand = union (enum) {
	instance_2d: Instance,
	post_process: PostProcess,

	pub const Instance = struct {
		texture: gfx.Handle,
		mesh_2d: gfx.Handle,
		transform: gfx.Transform2D,
	};

	pub const PostProcess = struct {

	};

	fn clone(self: RenderCommand, arena: *std.heap.ArenaAllocator) std.mem.Allocator.Error!RenderCommand {
		_ = arena;

		return switch (self) {
			.instance_2d => |instance_2d| .{.instance_2d = instance_2d},
			.post_process => |post_process| .{.post_process = post_process},
		};
	}
};

pub const RenderChain = commands.Chain(RenderCommand, RenderCommand.clone);

pub const RenderList = commands.List(RenderCommand, RenderCommand.clone);
