const App = @import("./App.zig");

const coral = @import("coral");

const std = @import("std");

fn Channel(comptime Message: type) type {
	return struct {
		buffers: [2]coral.stack.Sequential(Message),
		swap_index: u1 = 0,
		ticks: u1 = 0,

		const Self = @This();

		fn cleanup(channel: coral.Write(Self)) void {
			channel.res.deinit();
		}

		pub fn deinit(self: *Self) void {
			for (&self.buffers) |*buffer| {
				buffer.deinit();
			}

			self.* = undefined;
		}

		fn swap(channel: coral.Write(Self)) void {
			channel.res.ticks = coral.scalars.add(channel.res.ticks, 1) orelse 0;

			if (channel.res.ticks == 0) {
				channel.res.swap_index ^= 1;

				channel.res.buffers[channel.res.swap_index].clear();
			}
		}

		fn init(allocator: std.mem.Allocator) Self {
			return .{
				.buffers = .{
					.{.allocator = allocator},
					.{.allocator = allocator},
				},
			};
		}

		fn messages(self: Self) []const Message {
			return self.buffers[self.swap_index ^ 1].values;
		}

		fn push(self: *Self, message: Message) std.mem.Allocator.Error!void {
			try self.buffers[self.swap_index].push_grow(message);
		}
	};
}

pub fn Receive(comptime Message: type) type {
	const TypedChannel = Channel(Message);

	return struct {
		channel: *const TypedChannel,

		const Self = @This();

		pub const State = struct {
			channel: *const TypedChannel,
		};

		pub fn bind(context: coral.system.BindContext) std.mem.Allocator.Error!State {
			return .{
				.channel = (try context.register_read_only_resource_access(thread_restriction, TypedChannel)) orelse set: {
					try context.world.set_resource(thread_restriction, TypedChannel.init(coral.heap.allocator));

					break: set (try context.register_read_only_resource_access(thread_restriction, TypedChannel)).?;
				},
			};
		}

		pub fn init(state: *State) Self {
			return .{
				.channel = state.channel,
			};
		}

		pub fn messages(self: Self) []const Message {
			return self.channel.messages();
		}
	};
}

pub fn Send(comptime Message: type) type {
	const TypedChannel = Channel(Message);

	return struct {
		channel: *TypedChannel,

		const Self = @This();

		pub const State = struct {
			channel: *TypedChannel,
		};

		pub fn bind(context: coral.system.BindContext) std.mem.Allocator.Error!State {
			return .{
				.channel = (try context.register_read_write_resource_access(thread_restriction, TypedChannel)) orelse set: {
					try context.world.set_resource(thread_restriction, TypedChannel.init(coral.heap.allocator));

					const app = context.world.get_resource(.none, App).?;

					try context.world.on_event(app.events.post_update, coral.system_fn(TypedChannel.swap), .{
						.label = "swap channel of " ++ @typeName(Message),
					});

					try context.world.on_event(app.events.exit, coral.system_fn(TypedChannel.cleanup), .{
						.label = "clean up channel of " ++ @typeName(Message),
					});

					break: set (try context.register_read_write_resource_access(thread_restriction, TypedChannel)).?;
				},
			};
		}

		pub fn init(state: *State) Self {
			return .{
				.channel = state.channel,
			};
		}

		pub fn push(self: Self, message: Message) std.mem.Allocator.Error!void {
			try self.channel.push(message);
		}
	};
}

const thread_restriction = coral.World.ThreadRestriction.none;
