pub const App = @import("./App.zig");

pub const act = @import("./act.zig");

const coral = @import("coral");

const ext = @import("./ext.zig");

pub const gfx = @import("./gfx.zig");

pub const msg = @import("./msg.zig");

const std = @import("std");

pub const Options = struct {
	tick_rate: u64,
	execution: Execution,
	middlewares: []const *const Setup = default_middlewares,

	pub const Execution = union (enum) {
		single_threaded,
		thread_share: f32,
	};
};

pub const Setup = fn (*coral.World, App.Events) anyerror!void;

pub const default_middlewares = &.{
	gfx.setup,
	act.setup,
};

pub fn start_app(setup: Setup, options: Options) anyerror!void {
	defer {
		coral.heap.trace_leaks();
		ext.SDL_Quit();
	}

	var world = try switch (options.execution) {
		.single_threaded => coral.World.init(0),

		.thread_share => |thread_share| init: {
			const cpu_count = @as(u32, @intCast(std.math.clamp(std.Thread.getCpuCount() catch |cpu_count_error| {
				@panic(switch (cpu_count_error) {
					error.PermissionDenied => "permission denied retrieving CPU count",
					error.SystemResources => "system resources are preventing retrieval of the CPU count",
    				error.Unexpected => "unexpected failure retrieving CPU count",
				});
			}, 0, std.math.maxInt(u32))));

			break: init coral.World.init(coral.scalars.fractional(cpu_count, thread_share) orelse 0);
		},
	};

	defer world.deinit();

	const events = App.Events{
		.load = try world.create_event("load"),
		.pre_update = try world.create_event("pre-update"),
		.update = try world.create_event("update"),
		.post_update = try world.create_event("post-update"),
		.render = try world.create_event("render"),
		.finish = try world.create_event("finish"),
		.exit = try world.create_event("exit"),
	};

	const app = try world.set_get_resource(.none, App{
		.events = &events,
		.target_frame_time = 1.0 / @as(f64, @floatFromInt(options.tick_rate)),
		.is_running = true,
	});

	for (options.middlewares) |setup_middleware| {
		try setup_middleware(&world, events);
	}

	try setup(&world, events);
	try world.run_event(events.load);

	var ticks_previous = std.time.milliTimestamp();
	var accumulated_time = @as(f64, 0);

	while (app.is_running) {
		const ticks_current = std.time.milliTimestamp();
		const milliseconds_per_second = 1000.0;
		const delta_time = @as(f64, @floatFromInt(ticks_current - ticks_previous)) / milliseconds_per_second;

		ticks_previous = ticks_current;
		accumulated_time += delta_time;

		try world.run_event(events.pre_update);

		while (accumulated_time >= app.target_frame_time) : (accumulated_time -= app.target_frame_time) {
			try world.run_event(events.update);
		}

		try world.run_event(events.post_update);
		try world.run_event(events.render);
		try world.run_event(events.finish);
	}

	try world.run_event(events.exit);
}
